#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["quantize", "dump_config"]

import warnings
from collections import defaultdict

import numpy as np
import keras
from onnx import ModelProto
from keras.saving import get_registered_object, get_registered_name, serialize_keras_object

from .utils import apply_weights_to_model
from .transforms import sanitize
from .transforms.insert_layer import insert_in_config
from .transforms.transforms_utils import get_inbound_layers_config
from .calibrate import calibrate
from ..layers import (OutputQuantizer, WeightQuantizer, Dequantizer, quantization,
                      QuantizationParams, get_quantization_params, Attention, QuantizedConv2D,
                      QuantizedDepthwiseConv2D, QuantizedSeparableConv2D, QuantizedDense,
                      StatefulRecurrent, QuantizedActivation)
from ..layers.layers_base import (_GLOBAL_LAYER_TO_QLAYER, _GLOBAL_NO_OUTPUT_QUANTIZER,
                                  _GLOBAL_ALIGNED_INPUTS)
from ..onnx_support.quantization.quantize import quantize as quantize_onnx


# List of Quantizer layer's that do not have a float layer representation
NO_FLOAT_CUSTOM_QLAYERS = [Dequantizer, OutputQuantizer, WeightQuantizer]


def get_quantized_layer(layer):
    """ Returns the quantized version of the layer.

    Args:
        layer (keras.layers.Layer or dict): layer of interest

    Returns:
        keras.layer: quantized version of the layer if it exists, None otherwise.
    """
    if isinstance(layer, keras.layers.Layer):
        config = layer.get_config()
        layer_class = layer.__class__
    else:
        config = layer['config']
        layer_class = get_layer_class_from_config(layer)
    qlayer_class = _GLOBAL_LAYER_TO_QLAYER.get(layer_class.__name__, None)

    # Special case for activations: avoid quantization if activation is not within the
    # allowed activations
    if qlayer_class == QuantizedActivation:
        activation_name = get_registered_name(keras.activations.deserialize(config['activation']))
        if activation_name not in QuantizedActivation.arg_constraints['activation']():
            return None
    return qlayer_class


def is_quantized_layer(layer):
    """ Returns True when the layer is a quantized layer.

    Args:
        layer (keras.layers.Layer or type): layer of interest

    Returns:
        bool: True when the layer is a quantized layer, False otherwise.
    """
    if isinstance(layer, type):
        return layer in _GLOBAL_LAYER_TO_QLAYER.values()
    return layer.__class__ in _GLOBAL_LAYER_TO_QLAYER.values()


def get_layer_class_from_config(config):
    """ Returns the class object of a registered keras layer.

    Args:
        config (dict): the layer config

    Returns:
        type: the class of the layer
    """
    if hasattr(keras.layers, config["class_name"]):
        return getattr(keras.layers, config["class_name"])
    return get_registered_object(config["registered_name"])


def _handle_not_quantizable_layers(model, config, skip_warning=False):
    """ Includes a number of dequantizers such that the model is compatible.

    Args:
        model (keras.Model): model structure to check
        config (dict): config where Dequantizer(s) will be placed
        skip_warning (bool_optional): whether to skip warning provided by partial quantization.
            Defaults to False.

    Returns:
        bool: whether the model was fully quantized.
    """
    # Find where to insert the Dequantizer(s):
    # A dequantizer will be added for all links that connect a quantized layer to a floating one.
    num_no_quantizable_layers = 0
    dequantizer_inbounds = defaultdict(list)
    for layer in config['layers']:
        layer_class = get_layer_class_from_config(layer)

        if not (is_quantized_layer(layer_class) or layer_class in NO_FLOAT_CUSTOM_QLAYERS):
            # The current layer is float.
            for inbound_layer in get_inbound_layers_config(layer, config):
                inbound_layer_class = get_layer_class_from_config(inbound_layer)
                # A dequantizer will be added for each quantized inbound layer
                if is_quantized_layer(inbound_layer_class):
                    inbound_layer_name = inbound_layer['config']['name']
                    if len(dequantizer_inbounds) == 0 and not skip_warning:
                        # Print warning for the first non-quantizable layer
                        warnings.warn(f"'{layer['config']['name']}' of type {layer['class_name']} "
                                      "is not supported to quantize, a Dequantizer is added "
                                      "before it and quantization will stop at this layer.")
                    dequantizer_inbounds[inbound_layer_name].append(layer['config']['name'])
            num_no_quantizable_layers += 1

    if len(config['layers']) == num_no_quantizable_layers:
        raise RuntimeError(f"Impossible to quantize '{model.name}'. "
                           "At least one layer should be quantizable.")
    if len(dequantizer_inbounds) == 0:
        # Model was completely quantized.
        return True

    # Insert a Dequantizer on each target_layer -> outbound link
    for target_layer_name, outbound_names in dequantizer_inbounds.items():
        dequantizer = Dequantizer(name=f'{target_layer_name}/dequantizer')
        insert_in_config(model, target_layer_name, dequantizer, config, outbound_names)
    return False


def _prepare_output_quantizers(model):
    """ Parse the model and prepare OutputQuantizer configurations for layers requiring them.

    To ensure that an OutputQuantizer will be added to the latest possible layer in a 'block', the
    model is parsed in reverse order. If a layer requires aligned inputs, the function will find the
    preceding layer that can accept an OutputQuantizer and set it in the returned dictionary.

    Args:
        model (keras.Model): the model to parse

    Returns:
        dict: dictionary mapping layer names to an OutputQuantizer config.
    """
    # Dictionary that will contain layers and their OutputQuantizer configurations
    out_quantizer_configs = {}

    # Get quantization parameters
    qparams = get_quantization_params()

    def set_output_quantizer(layer_names, next_layer):
        """ Populates `out_quantizer_configs` with layer names and their OutputQuantizer. """
        for name in layer_names:
            current_layer = model.get_layer(name)
            # Handle special cases where the OutputQuantizer must be per-tensor:
            # - when current_layer has vector outputs,
            # - when next_layer is an Attention layer and the layer is Query or Key
            #   (first and second inputs)
            # - when the layer is a StatefulRecurrent layer
            if isinstance(current_layer, Attention):
                output_shape = current_layer.output_shape[0]
            else:
                output_shape = current_layer.output_shape
            # remove batch_size dim
            output_shape = output_shape[1:]
            vector_outputs = np.prod(output_shape) == output_shape[-1]
            query_or_key = (isinstance(current_layer, keras.layers.Dense)
                            and isinstance(next_layer, Attention)
                            and next_layer.inbound_nodes[0].inbound_layers.
                            index(current_layer) in [0, 1])
            is_stateful_rec = isinstance(current_layer, StatefulRecurrent)
            is_next_activation = get_quantized_layer(next_layer) == QuantizedActivation
            per_tensor = (query_or_key or vector_outputs or is_stateful_rec or
                          is_next_activation or qparams.per_tensor_activations)

            # If this is a new entry, set a default configuration
            if name not in out_quantizer_configs:
                axis = "per-tensor" if per_tensor else "per-axis"
                if isinstance(current_layer, keras.layers.ReLU):
                    params = dict(bitwidth=qparams.activation_bits,
                                  signed=qparams.activation_bits >= 8,
                                  axis=axis)
                elif isinstance(current_layer, keras.layers.Activation):
                    params = dict(bitwidth=qparams.activation_bits, axis=axis)
                elif is_next_activation:
                    params = dict(bitwidth=QuantizedActivation.DEFAULT_INPUT_BITWIDTH, axis=axis)
                else:
                    # StatefulRecurrent special: previous and self OutputQuantizer should be 16-bits
                    if is_stateful_rec or isinstance(next_layer, StatefulRecurrent):
                        bitwidth = 16
                    else:
                        bitwidth = qparams.output_bits
                    params = dict(bitwidth=bitwidth, axis=axis)
                params['buffer_bitwidth'] = qparams.buffer_bits
                out_quantizer_configs[name] = dict(output_quantizer=params)

            # If the layer OutputQuantizer configuration is already set, simply check the axis:
            # override the config if the outputs must be per-tensor
            else:
                current_axis = out_quantizer_configs[name]["output_quantizer"]["axis"]
                per_tensor = per_tensor or current_axis == "per-tensor"
                axis = "per-tensor" if per_tensor else "per-axis"
                out_quantizer_configs[name]["output_quantizer"]["axis"] = axis

    def cannot_have_output_quantizer(layer):
        """ Returns True when the layer cannot have an OutputQuantizer. """
        qlayer = get_quantized_layer(layer)
        return (isinstance(layer, Dequantizer)
                or qlayer is None
                or qlayer in _GLOBAL_NO_OUTPUT_QUANTIZER)

    def get_preceding_layer_names(layer):
        """ Retrieve inbounds layers names where an OutputQuantizer can be set. """
        previous_layers = []
        inbounds = layer.inbound_nodes[0].inbound_layers
        if not isinstance(inbounds, list):
            inbounds = [inbounds]
        for inbound in inbounds:
            # Skip input layers
            if isinstance(inbound, keras.layers.InputLayer):
                continue
            # When the given layer cannot have an OutputQuantizer, recursively call the function on
            # this layer
            if cannot_have_output_quantizer(inbound):
                previous_layers.extend(get_preceding_layer_names(inbound))
            else:
                previous_layers.append(inbound.name)
        return previous_layers

    # Parse the layers in reverse order
    for layer in model.layers[::-1]:
        # Find layers that will need aligned inputs
        if get_quantized_layer(layer) in _GLOBAL_ALIGNED_INPUTS:
            # Retrieve the inbounds that can have an OutputQuantizer
            previous_layers = get_preceding_layer_names(layer)
            # Set an OutputQuantizer in their inbounds
            set_output_quantizer(previous_layers, layer)

    return out_quantizer_configs


def quantize_keras(model, q_config=None, qparams=QuantizationParams(), samples=None,
                   num_samples=1024, batch_size=None, epochs=1, quantize_until=None):
    """Quantizes a Keras model using the provided configuration or parameters.

    Details on how this function behaves:

    - `q_config` has priority over `qparams`, meaning that when a match is found in `q_config` the
      given configuration will be used instead of `qparams`. This is useful to handle specific cases
      (e.g per-tensor output quantizer).
    - when no configuration is given, quantization parameters are deduced from `qparams` and
      OutputQuantizers are automatically set on appropriate layers.
    - `qparams` are only applied to 'float' Keras layers when they are first quantized. As a result,
      when re-quantizing a model, one must provide a complete `q_config`. This is made easy with the
      `dump_config` helper.

    If not already present, a final Dequantizer will be added at the end of the Model.

    The model will also be calibrated using the provided (or randomly generated inputs).

    Args:
        model (keras.Model): the model to quantize
        q_config (dict, optional): quantization configuration as a dictionary mapping layer names to
            their quantization configuration. Defaults to None.
        qparams (QuantizationParams, optional): global quantization parameters. Defaults to
            QuantizationParams().
        samples (tf.Dataset, np.array or generator, optional): calibration samples. When no samples
            are provided, random samples are generated. Defaults to None.
        num_samples (int, optional): number of samples to use in the provided samples or number of
            samples to generate. Defaults to 1024.
        batch_size (int, optional): the batch size. Defaults to None.
        epochs (int, optional): the number of epochs. Defaults to 1.
        quantize_until (str, optional): name of the layer until which to quantize:
            other layers after it will stay unchanged. Defaults to None.

    Returns:
        keras.Model: the quantized model
    """
    # Prevent requantization
    if any(is_quantized_layer(layer) for layer in model.layers):
        raise ValueError("Requantizing a model is not supported. "
                         "Please quantize the original float model directly.")

    q_config = q_config or dict()
    if quantize_until and not any(ly.name == quantize_until for ly in model.layers):
        raise ValueError(f"'{quantize_until}' is not a recognized layer in {model.name}")

    # Handle input_weight_bits using another QuantizationParams where
    # weight_bits = qparams.input_weight_bits, it will be set to False once the input layer has been
    # quantized.
    input_qparams = QuantizationParams(activation_bits=qparams.activation_bits,
                                       per_tensor_activations=qparams.per_tensor_activations,
                                       weight_bits=qparams.input_weight_bits,
                                       output_bits=qparams.output_bits,
                                       buffer_bits=qparams.buffer_bits)

    def get_quantize_layer(layer, quantize_config=None):
        """Get quantize config from float layer:
            - first, we get its quantized version,
            - then, we return the quantized layer with config updated
        """
        # Check if qlayer exists in custom layers and returns the float version of the layer if not
        l_class = get_layer_class_from_config(layer)
        ql_class = get_quantized_layer(layer)
        if ql_class is None:
            ql_class = l_class

        # Initialize quantized layer from the float config
        qlayer = layer

        # Instantiate quantized layer from configuration if there is one
        if quantize_config:
            qlayer['config']['quant_config'] = quantize_config
        # Set the preset default configuration otherwise
        elif qlayer['config']['name'] in out_quantizer_configs:
            qlayer['config']['quant_config'] = out_quantizer_configs[qlayer['config']['name']]

        # Retrieve the quantized config after initializing the quantized layer, in order to
        # configure the specific parameters given by the QuantizationParams context.
        new_layer = ql_class.from_config(qlayer['config'])
        qlayer.update(serialize_keras_object(new_layer))
        return qlayer

    # Sanitize the model and make it quantization ready
    model = sanitize(model)

    # Determine where to set OutputQuantizers, the return dict will be used as a non-local
    # variable in the _replace_layer function.
    with quantization(qparams):
        out_quantizer_configs = _prepare_output_quantizers(model)

    # Quantize the model, modifying each layer config by its respective quantized version
    input_layers = (QuantizedConv2D, QuantizedDepthwiseConv2D, QuantizedSeparableConv2D,
                    QuantizedDense)
    qmodel_config = model.get_config()
    quantized_layers = set()
    for idx, layer in enumerate(qmodel_config['layers']):
        # Retrieve quantize config from layer
        match_conf = q_config.get(layer['config']['name'], None)

        # Overwrite quantization context with input_qparams (if they are not None)
        with quantization(input_qparams or qparams):
            inbound_layers = get_inbound_layers_config(layer, qmodel_config)
            # Quantization is only performed if the inbound layers were quantized
            if all(x['config']['name'] in quantized_layers for x in inbound_layers):
                qlayer = get_quantize_layer(layer, match_conf)
            else:
                qlayer = layer

        # When the qlayer is an input layer that has been quantized, disable input_qparams
        qlayer_class = get_layer_class_from_config(qlayer)
        if input_qparams and qlayer_class in input_layers:
            input_qparams = None

        # Skip input layers
        if qlayer_class == keras.layers.InputLayer:
            # Although InputLayer is not quantizable, layer is treated as one
            # so its outbounds can be quantized.
            quantized_layers.add(qlayer['config']['name'])
            continue

        # If it was not possible to quantize the layer, try to quantize the next one.
        # This ensures that as many layers as possible are quantized.
        if not is_quantized_layer(qlayer_class):
            continue

        # Finally, update model with quantize layer config
        # Note at this point, we know the layer was quantized successfully
        qmodel_config['layers'][idx] = qlayer
        if quantize_until != layer["config"]["name"]:
            # If quantize_until is provided, layer is quantized but is not added to
            # quantized_layers list, preventing the quantization of layers after it.
            # Note if layer is within a branch, quantization will end only for this branch
            quantized_layers.add(qlayer['config']['name'])

    # Insert the number of Dequantizers necessary for the model to be compatible
    is_full_quantized = _handle_not_quantizable_layers(model,
                                                       qmodel_config,
                                                       skip_warning=quantize_until is not None)

    # Build the model and transfer weights
    qmodel = model.from_config(qmodel_config)
    apply_weights_to_model(qmodel, {var.name: var for var in model.variables}, False)

    # Convert model into a functional one.
    # Note if model was completely quantized, we add a last dequantizer to produce a float output
    y = qmodel.output
    if is_full_quantized:
        y = Dequantizer()(y)
    qmodel = keras.Model(qmodel.input, y, name=model.name)

    # Now that the model is quantized, proceed to calibration
    calibrate(model, qmodel, samples=samples, num_samples=num_samples, batch_size=batch_size,
              epochs=epochs)
    return qmodel


def quantize(model, q_config=None, qparams=QuantizationParams(), samples=None, num_samples=1024,
             batch_size=None, epochs=1, quantize_until=None):
    """Quantizes a Keras or ONNX model using the provided configuration or parameters.

    Details on how this function behaves:

    - `q_config` has priority over `qparams`, meaning that when a match is found in `q_config` the
      given configuration will be used instead of `qparams`. This is useful to handle specific cases
      (e.g per-tensor output quantizer). This is only used when quantizing Keras models.
    - when no configuration is given, quantization parameters are deduced from `qparams` and
      OutputQuantizers are automatically set on appropriate layers.
    - `qparams` are only applied to 'float' Keras layers when they are first quantized. As a result,
      when re-quantizing a model, one must provide a complete `q_config`. This is made easy with the
      `dump_config` helper. Note the only configuration supported when quantizing ONNX models is
      8-bit for weights and activations, but per_tensor_activations param will be taken into
      account.

    If not already present, a final Dequantizer will be added at the end of the Model.

    The model will also be calibrated using the provided (or randomly generated inputs).

    Args:
        model (keras.Model or ModelProto): the model to quantize
        q_config (dict, optional): quantization configuration as a dictionary mapping layer names to
            their quantization configuration. Defaults to None.
        qparams (QuantizationParams, optional): global quantization parameters. Defaults to
            QuantizationParams().
        samples (tf.Dataset, np.array or generator, optional): calibration samples. When no samples
            are provided, random samples are generated. Defaults to None.
        num_samples (int, optional): number of samples to use in the provided samples or number of
            samples to generate. Defaults to 1024.
        batch_size (int, optional): the batch size. Defaults to None.
        epochs (int, optional): the number of epochs. This parameter must be 1 for ONNX models.
            Defaults to 1.
        quantize_until (str, optional): name of the layer/node until which to quantize:
            other ones after it will stay unchanged. Defaults to None.

    Returns:
        keras.Model or ModelProto: the quantized model
    """
    # Calibration with random samples will only provide meaningful results when quantizing
    # per-tensor
    if samples is None and not qparams.per_tensor_activations:
        warnings.warn("Quantizing per-axis with random calibration samples is not accurate.\
                       Set QuantizationParams.per_tensor_activations=True when calibrating with \
                       random samples.")
    if type(model) != ModelProto:
        return quantize_keras(model=model,
                              q_config=q_config,
                              qparams=qparams,
                              samples=samples,
                              num_samples=num_samples,
                              batch_size=batch_size,
                              epochs=epochs,
                              quantize_until=quantize_until)
    elif q_config:
        raise ValueError("unsupported parameter q_config for ONNX models quantization")
    elif epochs != 1:
        raise ValueError("unsupported parameter epochs != 1 for ONNX models quantization")
    return quantize_onnx(model=model,
                         qparams=qparams,
                         samples=samples,
                         num_samples=num_samples,
                         batch_size=batch_size,
                         quantize_until=quantize_until)


def dump_config(model):
    """Dump the quantization configuration of a quantized model, exporting the configuration for
    each quantized layer.

    Args:
        model (keras.Model): a quantized model.

    Returns:
        dict: the configuration of the model.
    """
    # Get the configuration of the model, iterating over each layer and updating on config.
    config = {}
    for layer in model.layers:
        # Try to take the current quantized configuration
        ly_config = layer.get_config().get('quant_config')

        # Only append quantized configuration
        if is_quantized_layer(layer) and ly_config:
            config[layer.name] = ly_config

    return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["generate_keras_random_samples", "generate_onnx_random_samples"]


import numpy as np
import tensorflow as tf
import copy

from ...tensors import QFloat, FixedPoint
from ...models import get_model_input_dtype
from ...onnx_support.graph_tools import value_info_to_tensor_shape


def _get_np_deterministic(size, dtype, min_value=None, max_value=None, rng=None):
    rng = rng or np.random.default_rng()
    dtype = np.dtype(dtype)
    if np.issubdtype(dtype, np.integer):
        iinfo = np.iinfo(dtype)
        min_value = iinfo.min if min_value is None else min_value
        max_value = iinfo.max if max_value is None else max_value
        x = rng.integers(min_value, max_value, size, endpoint=True)
    else:
        min_value = 0.0 if min_value is None else min_value
        max_value = 1.0 if max_value is None else max_value
        x = rng.uniform(min_value, max_value, size)
    return x.astype(dtype)


def _get_tf_deterministic(tf_spec, batch_size=1, min_value=None, max_value=None, rng=None):
    assert isinstance(tf_spec, tf.TensorSpec)
    input_shape = (batch_size, *tf_spec.shape[1:])
    x = _get_np_deterministic(input_shape, tf_spec.dtype.as_numpy_dtype, min_value, max_value, rng)
    return tf.cast(x, tf_spec.dtype)


def _update_keras_spec(keras_spec, value_bits=None, dtype=None):
    def _replace_fp_spec(fp_spec, value_bits):
        if value_bits is not None:
            fp_spec.__dict__["value_bits"] = value_bits

    keras_spec = copy.deepcopy(keras_spec)
    if isinstance(keras_spec, FixedPoint.Spec):
        _replace_fp_spec(keras_spec, value_bits)
    elif isinstance(keras_spec, QFloat.Spec):
        _replace_fp_spec(keras_spec.fp, value_bits)
    elif isinstance(keras_spec, tf.TensorSpec) and dtype is not None:
        keras_spec._dtype = tf.as_dtype(dtype)
    return keras_spec


def generate_keras_random_samples(model, batch_size=1, seed=None):
    """Generate a random set of inputs for a model.

    Args:
        model (tf.keras.Model): the target model to generate inputs.
        batch_size (int, optional): a batch size. Defaults to 1.
        seed (int, optional): a random seed (reproducibility purpose). Defaults to None.

    Returns:
        tf.Tensor: a set of samples
    """
    rng = np.random.default_rng(seed=seed)
    input_dtype = get_model_input_dtype(model)
    input_specs_list = model.input if isinstance(model.input, (list, tuple)) else [model.input]
    gen_inputs = []
    for idx, keras_spec in enumerate(input_specs_list):
        if seed is not None:
            rng = np.random.default_rng(seed=seed + idx)
        # Replace dtype in spec to generate integer values
        integer_keras_spec = _update_keras_spec(keras_spec.type_spec, dtype=input_dtype)
        xq = _get_tf_deterministic(integer_keras_spec, batch_size, rng=rng)
        gen_inputs.append(tf.cast(xq, keras_spec.dtype))

    if not isinstance(model.input, (list, tuple)):
        return gen_inputs[0]
    return gen_inputs


def generate_onnx_random_samples(model, batch_size=1, seed=None):
    """Generate a random set of inputs for a model.

    Args:
        model (ONNXModel): the target model to generate inputs.
        batch_size (int, optional): a batch size. Defaults to 1.
        seed (int, optional): a random seed (reproducibility purpose). Defaults to None.

    Returns:
        np.ndarray: a set of samples
    """
    rng = np.random.default_rng(seed=seed)
    gen_inputs = {}
    for idx, input_vi in enumerate(model.input):
        if seed is not None:
            rng = np.random.default_rng(seed=seed + idx)

        # Retrieve shape from value info
        shape, dtype = value_info_to_tensor_shape(input_vi)
        gen_inputs[input_vi.name] = _get_np_deterministic((batch_size, *shape[1:]), dtype, rng=rng)
    return gen_inputs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from copy import copy
import tensorflow as tf


from ..debugging import assert_equal


@tf.custom_gradient
def round_through(x):
    rounded = tf.math.round(x)

    def grad(upstream):
        return upstream
    return rounded, grad


@tf.custom_gradient
def floor_through(x):
    floored = tf.math.floor(x)

    def grad(upstream):
        return upstream
    return floored, grad


@tf.custom_gradient
def ceil_through(x):
    ceiled = tf.math.ceil(x)

    def grad(upstream):
        return upstream
    return ceiled, grad


@tf.custom_gradient
def saturate(x, n):
    """Saturate x values on [-2^n, 2^n - 1]

    This function is a wrapper around tf.clip_by_values that exhibits a
    simplified gradient for a better efficiency.
    Unlike tf.clip_by_value, the gradient outside the clipping range is not zero.

    Args:
        x (tf.Tensor): a tensor of float values representing integer.
        n (tf.Tensor): the target bitwidth.

    Returns:
        the resulting clipped tensor and its gradient function.

    """
    # Clamp values to [-2^n, 2^n - 1]
    int_max = tf.math.pow(2.0, n)
    saturated = tf.clip_by_value(x, -int_max, int_max - 1, name="saturate")

    def grad(upstream):
        # The gradient of x is 1 everywhere
        dx = upstream
        # The gradient of n is zero (constant)
        dn = tf.zeros_like(n)
        return dx, dn

    return saturated, grad


class QTensor(tf.experimental.ExtensionType):
    """Abstract class to exchange quantized tensors between layers
    """
    shape: tf.TensorShape  # Required to convert to a KerasTensor

    @property
    def name(self):
        """Returns the QTensor name

        Returns:
            str: the QTensor name
        """
        raise NotImplementedError

    @property
    def per_tensor(self):
        """Returns if QTensor is quantized per-tensor

        Returns:
            bool: True if QTensor is quantized per-tensor or False on per-axis case.
        """
        raise NotImplementedError

    def to_float(self):
        """Returns a float representation of the QTensor

        Returns:
            tf.Tensor: the float representation.
        """
        raise NotImplementedError

    def __getitem__(self):
        """Retrieve a slice or element from the QTensor

        Returns:
            :obj:`QTensor`: the sliced QTensor
        """
        raise NotImplementedError

    def clone(self):
        """Returns a copy of the QTensor

        Returns:
            :obj:`QTensor`: the copy.
        """
        return copy(self)

    def __str__(self):
        class_name = self.__class__.__name__
        x_float = self.to_float()
        return f"{class_name}: {x_float}"

    @staticmethod
    def int_max(value_bits):
        return 2 ** value_bits - 1

    def assert_per_tensor(self):
        """Asserts that a QTensor is quantized per-tensor"""
        assert_equal(self.per_tensor, True, message=f"{self.name} is not per-tensor.")


def pow2(n):
    """Return the power of two of an integer

    Args:
        n (`tf.tensor`, int): the positive or negative exponent

    Returns:
        tf.Tensor: a float tensor containing the PoT of the input.
    """
    return tf.pow(2.0, tf.cast(n, tf.float32))


def ceil_log2(x):
    """Return the closest power of two exponent of a float tensor.

    This evaluates for each element of the input tensor the integer exponent leading
    to the closest power-of-two higher than the input.

    In hardware, if the inputs are represented as integer, this operation can
    be implemented by identifying the leading bit index and increment the result
    by 1.

    Example: ceil_log2(7) = ceil_log2(0b00000111) = 2 + 1 = 3

    Args:
        x (tf.Tensor): the source tensor

    Returns:
        tf.Tensor: a float tensor containing integer values
            representing the closest PoT exponents.
    """
    return ceil_through(tf.experimental.numpy.log2(tf.cast(x, tf.float32)))


def floor_log2(x):
    """Return the closest power of two exponent of a float tensor.

    This evaluates for each element of the input tensor the integer exponent leading
    to the closest power-of-two lower than the input (except if x < 1).

    In hardware, if the inputs are represented as integer, this operation can
    be implemented by identifying the leading bit index.

    Example: floor_log2(7) = floor_pow2(0b00000111) = 2

    Args:
        x (tf.Tensor): the source tensor

    Returns:
        tf.Tensor: a float tensor containing integer values
            representing the closest PoT exponents.
    """
    return floor_through(tf.experimental.numpy.log2(tf.cast(x, tf.float32)))


def round_log2(x):
    """Return the closest power of two exponent of a float tensor.

    This evaluates for each element of the input tensor the integer exponent leading
    to the closest power-of-two.

    In hardware, if the inputs are represented as integer, this operation can
    be implemented by:
    - identifying the leading bit index,
    - increment by 1 if the previous bit is 1 also.

    Example: round_log2(7) = round_log2(0b00000111) = 2 + 1 = 3
             round_log2(5) = round_log2(0b00000101) = 2

    Args:
        x (tf.Tensor): the source tensor

    Returns:
        tf.Tensor: a float tensor containing integer values
            representing the closest PoT exponents.
    """
    return round_through(tf.experimental.numpy.log2(tf.cast(x, tf.float32)))


def convert_ellipsis_into_slices(values_rank, idx):
    """
    Converts an Ellipsis (`...`) if found in the indexing tuple into a series of slice objects
    that cover the remaining dimensions of the tensor.

    Args:
        values_rank (int): The rank (number of dimensions) of the tensor being indexed.
        idx (tuple): The indexing tuple that may contain an Ellipsis.

    Returns:
        tuple: A new indexing tuple with the Ellipsis replaced by appropriate slices.
    """
    new_idx = ()
    for element in idx:
        if element == ...:
            for _ in range(values_rank - len(idx) + 1):
                new_idx += (slice(None, None, None),)
        else:
            new_idx += (element,)
    return new_idx


def slice_tensor_by_index(idx, values_rank, tensor):
    """
    Slices the frac_bits or scales tensor when slicing a QTensor.

    Args:
        idx (tuple/slice/int/Ellipsis): The indexing tuple.
        values_rank (int): The rank (number of dimensions) of the values tensor.
        tensor (tf.Tensor): The frac_bits/scale tensor to slice.

    Returns:
        tf.Tensor: the sliced frac_bits/scale tensor tensor.
    """

    tensor_rank = tf.rank(tensor)
    if isinstance(idx, tuple):
        # Convert Ellipsis into actual slices
        if any(x is Ellipsis for x in idx):
            idx = convert_ellipsis_into_slices(values_rank, idx)

        # Case where the number of indices matches the rank of the values tensor
        if values_rank == len(idx):
            new_fb_tuple = idx[-tensor_rank:]
            return tensor[new_fb_tuple]

        # Case where the frac_bits/scales tensor has fewer dimensions than the remaining positions
        elif values_rank - len(idx) - tensor_rank >= 0:
            return tensor

        # Case where the frac_bits/scales tensor needs more dimensions
        else:
            m = tensor_rank - (values_rank - len(idx))
            fb_ind = idx[-m:]
            fb_ind += (Ellipsis,)
            return tensor[fb_ind]

    # Case where the index is not a tuple (int, ellipsis, scalar)
    # and values values has more dimensions than the frac_bits/scales tensor
    elif values_rank > tensor_rank:
        return tensor

    # Case where the values and tensor have the same dimensions
    else:
        return tensor[idx]



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["QuantizedDense1D", "get_qgemm"]

import numpy as np
from onnx import TensorProto as TP
from onnx.helper import make_node

from .base_layer import OnnxLayer
from .subgraph_ops import cast_tensors_to, get_scale_out_ops
from .subgraph_ops.activation import get_activation_ops
from .layer_compatibility import check_clip_relu_compatibility

from ..graph_tools import (TENSOR_SHAPE, get_node, get_variable, get_tensor_shape,
                           check_node_attributes)
from ..quantization.weights import quantize_weights, quantize_vector, align_to
from ..quantization.outputs import downscale


def get_qgemm(nodes, graph):
    gemm_node = get_node(nodes, 'Gemm')
    assert gemm_node is not None

    # Check supported attributes
    check_node_attributes(gemm_node, {'alpha': [1.0], 'beta': [1.0], 'transA': [0], 'transB': [1]})

    # Retrieve attributes
    flatten = bool(get_node(nodes, 'Flatten'))
    act_node = get_node(nodes, 'Relu')
    clip_node = get_node(nodes, 'Clip')
    qgemm = QuantizedDense1D(flatten=flatten,
                             activation=bool(act_node) or bool(clip_node),
                             name=gemm_node.name)

    # Sets the weights to configure the operation chain
    qgemm.set_weight("kernel", get_variable(gemm_node.input[1], graph))
    # If third attribute is there and it is not empty, then there is a bias
    if len(gemm_node.input) == 3 and gemm_node.input[2]:
        qgemm.set_weight("bias", get_variable(gemm_node.input[2], graph))
    if clip_node:
        check_clip_relu_compatibility(clip_node, graph)
        qgemm.set_weight("max_value", get_variable(clip_node.input[2], graph))

    return qgemm


class QuantizedDense1D(OnnxLayer):
    """Intermediate representation of Flatten() + QGemm() + ReLU() as an exportable node.

    Args:
        flatten (bool, optional): whether to flatten the inputs. Defaults to False.
        activation (bool, optional): whether to apply relu operation. Defaults to False.
        name (str, optional): the node name. Defaults to ''.
    """

    def __init__(self, flatten=False, activation=False, name=''):
        super().__init__("QuantizedDense1D", name=name)

        # Save properties need to serialize operation name
        self.serialize_attr["flatten"] = flatten
        self.serialize_attr["activation"] = activation

        # Declare weights
        self._add_weight("kernel")
        self._add_weight("bias")
        self._add_weight("max_value")

    def __build__(self, input_ts, downscale=True):
        assert input_ts.dtype == np.int8
        assert self.weights["kernel"].ndim == 2
        filters = self.weights["kernel"].shape[0]

        # The chain of operations is modified if downscale is needed
        self.serialize_attr["scale"] = downscale

        # Compute output shape
        output_type = "int8" if downscale else "int32"
        output_ts = TENSOR_SHAPE((input_ts.shape[0], filters), np.dtype(output_type))
        return output_ts

    def __quantize__(self, qinput, out_tensor_range, force_fp=False):
        i_scale = qinput.weights["scale"]
        kernel = self.weights["kernel"]
        filters, channels = kernel.shape

        # Rescale kernel according to input scale. This operation is different if
        # pattern contain a Flatten.
        assert i_scale.ndim <= 1
        if 'Flatten' in self.op_type:
            # If flatten is there, we need to reshape weights to apply input scale
            _, c, x, y = get_tensor_shape(self.input)
            # Unroll first flattened inputs
            kernel = np.reshape(kernel, (filters, c, x, y))
            # Divide kernel by input shape (that has shape of c)
            kernel = kernel / align_to(i_scale, kernel.ndim)
            # Reshape back to original shape
            kernel = np.reshape(kernel, (filters, channels))
        else:
            kernel = kernel / align_to(i_scale, kernel.ndim)
        # Quantize and set weights
        qweights, i_scale = quantize_weights(kernel)

        # Prepare tensors list with unique names
        gemm_name = self.name
        prefix = gemm_name + "_"
        weights_dict = {prefix + "Wi": qweights}
        if "Biased" in self.op_type:
            qbias = quantize_vector(self.weights["bias"], i_scale)
            weights_dict[prefix + "B"] = qbias

        # Reshape i_scale to match with channel axis
        i_scale = align_to(i_scale, qweights.ndim)

        # Quantize max value when there is an activation
        if "Clipped" in self.op_type:
            qmax_value = quantize_vector(self.weights["max_value"], i_scale, signed=False)
            weights_dict[prefix + "max_value"] = qmax_value

        if "Scaled" not in self.op_type:
            output_scale = i_scale.squeeze()
        else:
            # Now consider calibrated output range
            scale, s_out, output_scale = downscale(out_tensor_range, i_scale, force_fp=force_fp)
            # Add scale out inputs and weights
            weights_dict[prefix + "M"] = scale.astype("uint8")
            weights_dict[prefix + "S_out"] = s_out

        # Return quantized weights and ouput scale
        return weights_dict, output_scale

    @staticmethod
    def build_subgraph(op_type):
        # Cast input, weights (and bias) into float.
        t_names = ["X", "W", ""]
        if "Biased" in op_type:
            t_names[-1] = "bias"
        nodes, t_names = cast_tensors_to(t_names)

        # Flatten (optional)
        if "Flatten" in op_type:
            nodes.append(make_node("Flatten", inputs=t_names[:1], outputs=["Xflat"]))
            t_names[0] = "Xflat"

        # Gemm
        nodes.append(make_node("Gemm", inputs=t_names, outputs=["Yi"], transB=1))

        # Activation (optional)
        if "ReLU" in op_type:
            # Replace previous output as relu input
            nodes[-1].output.__setitem__(0, nodes[-1].op_type)
            nodes += get_activation_ops(nodes[-1].output[0], "Yi", "ReLUClipped" in op_type)

        # Apply final scale (with saturation) (optional)
        if "Scaled" in op_type:
            shift_nodes, shift_t_names = cast_tensors_to(["Scale", "Shift"])
            nodes += shift_nodes
            nodes += get_scale_out_ops("Yi", "Yscaled", *shift_t_names, saturate=True)
            nodes.append(make_node("Cast", ["Yscaled"], ["Y"], to=TP.INT8))
        else:
            nodes.append(make_node("Cast", ["Yi"], ["Y"], to=TP.INT32))
        return nodes



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np


def to_fixed_point(x, bitwidth, signed=True, clamp=False):
    """Convert a number to a FixedPoint representation

    The representation is composed of a mantissa and an implicit exponent expressed as
    a number of fractional bits, so that:

    x ~= mantissa . 2 ** -frac_bits

    The mantissa is an integer whose bitwidth and signedness are specified as parameters.

    Args:
        x (np.ndarray): the source number or array
        bitwidth (np.ndarray): the desired bitwidth
        signed (bool, optional): when reserving a bit for the sign. Defaults to True.
        clamp (bool, optional): clamp the resulted fractional bits to mantissa_bits.
            Defaults to False.

    Returns:
        np.ndarray, np.ndarray: the mantissa and fractional bits
    """
    if not isinstance(x, np.ndarray):
        x = np.array(x)
    if np.any(np.isinf(x)):
        raise ValueError(f"Infinite values are not supported. Receives: {x}")
    # Evaluate the number of bits available for the mantissa
    mantissa_bits = bitwidth - 1 if signed else bitwidth
    # Evaluate the number of bits required to represent the whole part of x
    # as the power of two enclosing the absolute value of x
    # Note that it can be negative if x < 0.5, as well as we force whole_bits = 0 when x is 0
    y = np.abs(np.where(x == 0, 1, x))
    whole_bits = np.ceil(np.log2(y)).astype(np.int32)
    # Deduce the number of bits required for the fractional part of x
    # Note that it can be negative if the whole part exceeds the mantissa
    frac_bits = mantissa_bits - whole_bits
    if clamp:
        frac_bits = np.minimum(frac_bits, mantissa_bits)
    # Evaluate the 'scale', which is the smallest value that can be represented (as 1)
    scale = 2. ** -frac_bits
    # Evaluate the minimum and maximum values for the mantissa
    mantissa_min = -2 ** mantissa_bits if signed else 0
    mantissa_max = 2 ** mantissa_bits - 1
    # Evaluate the mantissa by quantizing x with the scale, clipping to the min and max
    mantissa = np.clip(np.round(x / scale), mantissa_min, mantissa_max).astype(np.int32)
    return mantissa, frac_bits


def to_float(mantissa, frac_bits):
    return mantissa * 2. ** -frac_bits



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Helper that replaces lambdas with their equivalent Keras layer.
"""

__all__ = ["replace_lambda"]

from copy import deepcopy
from keras.layers import Activation
from keras.src.layers import TFOpLambda

from .transforms_utils import get_layers_by_type, get_layer_index, inbound_node_generator


def _update_inbound_nodes(layer_config, target):
    """ Update a Lambda layer inbound node towards their Layer equivalent.

    Args:
        layer_config (dict): config of the lambda layer
        target (str): class name of the target layer

    Returns:
        str: name of the lambda operation
    """
    name = None
    if 'inbound_nodes' not in layer_config:
        # A layer from a sequence model does not have 'inbound_nodes'.
        # So, nothing to do in this case
        return name
    for inbound_node in inbound_node_generator(layer_config):
        if isinstance(inbound_node, dict):
            inbound_node = inbound_node.values()
        connection_info = inbound_node[0]
        # "connection_info[-1]" holds the lambda config, that is the op name and other lambda
        # specific parameters. Code below will retrieve meaningful parameters that will be used to
        # define the config of the new layer and will then be dropped.
        if target == 'Reshape':
            # Set the 'target_shape' parameter from the 'shape' attribute
            layer_config['config']['target_shape'] = connection_info[-1].get('shape')
        elif target == 'Permute':
            # Set the 'dims' parameter from the 'perm' attribute
            perm = connection_info[-1].get('perm')
            # Permute 'dims' must start at 1 but transpose 'perm' starts at 0
            layer_config['config']['dims'] = [p + 1 for p in perm]
        elif target == "Add":
            # Get the second inbound currently defined as {'y': ['name', 0, 0]}
            other_inbound = connection_info[-1].get('y')
            # Modify it to fit the ['name', 0, 0, {}] convention
            other_inbound.append({})
            # Add the updated inbound into the inbound_node list and set it in layer_config
            inbound_node.append(other_inbound)
            layer_config['inbound_nodes'] = [inbound_node]
        # Retrieve the lambda name as it will be used to set the name for the layer
        name = connection_info[-1].get('name', None)
        # Drop the lambda config
        connection_info[-1] = {}
    return name


def replace_lambda(model):
    """ Replaces lambda layers from a model with their equivalent Keras layer.

    This transform handles the following replacements:

        - Lambda(relu) or Activation('relu') → ReLU,
        - Lambda(transpose) → Permute,
        - Lambda(reshape) → Reshape,
        - Lambda(add) → Add,
        - Lambda('gelu') → Activation('gelu'),
        - Lambda('silu') → Activation('silu').

    Args:
        model (keras.Model): the model of interest

    Returns:
        keras.Model: the original model or a new one with lambda replaced.
    """
    # Map function names to Keras layers
    lambda_to_layer = {
        'nn.relu': 'ReLU',
        'math.add': 'Add',
        '__operators__.add': 'Add',
        'reshape': 'Reshape',
        'transpose': 'Permute',
        'compat.v1.transpose': 'Permute',
        'nn.silu': 'Activation',
        'nn.gelu': 'Activation',
    }

    # Get all Activations and TFOpLambda layers present in the model
    lambdas = get_layers_by_type(model, (Activation, TFOpLambda))

    # When there are no valid candidates, return the original model
    if not lambdas:
        return model

    # Copy configuration before applying modifications
    config = deepcopy(model.get_config())

    for layer in lambdas:
        layer_index = get_layer_index(config['layers'], layer.name)
        layer_config = config['layers'][layer_index]
        # Replace 'relu' Activations layers with ReLU layers
        if (layer_config['class_name'] == 'Activation'
                and layer_config['config']['activation'] == 'relu'):
            # Drop the 'activation' parameter and update 'class_name'
            layer_config['config'].pop('activation')
            layer_config['class_name'] = 'ReLU'
        # Replace TFOpLambda layers
        elif layer_config['class_name'] == 'TFOpLambda':
            # Retrieve the function used in the config and get the equivalent Keras layer name
            target = lambda_to_layer.get(layer_config['config']['function'], None)
            if target:
                # Drop the 'function' parameter and update 'class_name'
                op_name = layer_config['config'].pop('function').replace('nn.', '')
                layer_config['class_name'] = target
                # If target is 'Activation', put op_name in 'activation' parameter
                if target == "Activation":
                    layer_config['config']['activation'] = op_name
                # Update the inbound part of the config: the last element of the inbound list of
                # lambda layers will contain the lambda op parameters that are used to set the
                # config for the new layer.
                new_name = _update_inbound_nodes(layer_config, target)
                # If layer name was updated, use the new name everywhere in the config
                if new_name:
                    # Serialize the dict into a string
                    str_config = str(config)
                    # Replace name using 'old_name' for an exact match
                    str_config = str_config.replace(f"'{layer_config['name']}'", f"'{new_name}'")
                    # Deserialize the updated string into a dict
                    config = eval(str_config)

    # Reconstruct model from the config
    updated_model = model.from_config(config)

    # Restore model weights
    updated_model.set_weights(model.get_weights())
    return updated_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["Quantizer", "Dequantizer"]

import tensorflow as tf
import keras

from keras.layers import Layer

from ...tensors import QTensor, QFloat
from ..recorders import TensorRecorder


class Quantizer(Layer):
    """The base class for all quantizers.

    The bitwidth defines the number of quantization levels on which the
    values will be quantized.
    For a quantizer that accepts unsigned values, the maximum quantization
    level is :math:`2 ^ {bitwidth} - 1`.
    For a quantizer that accepts signed values, we lose one bit of precision to
    store the sign.
    When the quantizer is signed, the quantization interval is asymmetric around
    zero (i.e range: :math:`[- 2 ^ {bitwidth - 1}, 2 ^ {bitwidth - 1} - 1]`).

    Args:
        bitwidth (int): the quantization bitwidth.
        signed (bool, optional): whether the quantizer expects signed values or unsigned.
            Defaults to True.
    """

    def __init__(self, bitwidth, signed=True, **kwargs):
        min_bitwidth = 2 if signed else 1
        if not isinstance(bitwidth, int) or bitwidth < min_bitwidth:
            raise ValueError(
                f"Bitwidth should be an int >= {min_bitwidth}, currently {bitwidth}")
        self.bitwidth = bitwidth
        self.signed = signed
        self.value_bits = bitwidth - 1 if signed else bitwidth
        super().__init__(**kwargs)

    def get_config(self):
        """Get the config of the layer.

        Returns:
            dict: the config of the layer.
        """
        config = super().get_config()
        config.update({"bitwidth": self.bitwidth})
        config.update({"signed": self.signed})
        return config


@keras.saving.register_keras_serializable()
class Dequantizer(Layer):
    """ Layer that allows to dequantize its inputs.
    """
    scales: list = None
    frac_bits: list = None

    def _build_records(self, inputs):
        def _build(x):
            record_fb = record_scale = None
            # From Keras documentation, any variable creation taking place
            # in call should be wrapped with tf.init_scope
            with tf.init_scope():
                if isinstance(x, QTensor):
                    record_fb = TensorRecorder(self.name + "/record_fb")
                if isinstance(x, QFloat):
                    record_scale = TensorRecorder(self.name + "/record_scale")
            return record_fb, record_scale

        if self.frac_bits is not None:
            # Nothing to do
            return
        if not isinstance(inputs, (tuple, list)):
            # Manage single inputs
            self.frac_bits, self.scales = _build(inputs)
            return

        self.frac_bits = []
        self.scales = []
        with tf.init_scope():
            for x in inputs:
                frac_bits, scales = _build(x)
                self.frac_bits.append(frac_bits)
                self.scales.append(scales)

    def call(self, inputs):
        """Convert QTensor inputs to float.

        Args:
            inputs (tf.Tensor or :obj:`QTensor`): the inputs tensor(s).

        Returns:
            tf.Tensor: the dequantized tensor(s).
        """

        def dequantize(x, frac_bits_recorder=None, scales_recorder=None):
            if isinstance(x, QTensor):
                if frac_bits_recorder is not None:
                    frac_bits_recorder(x.fp.frac_bits if isinstance(x, QFloat) else x.frac_bits)
                if scales_recorder is not None:
                    scales_recorder(x.scales)
                return x.to_float()
            return x

        # Build records
        self._build_records(inputs)

        # Apply dequantizer
        if isinstance(inputs, (list, tuple)):
            return [dequantize(x, fb, scales) for x, fb, scales in
                    zip(inputs, self.frac_bits, self.scales)]

        return dequantize(inputs, self.frac_bits, self.scales)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["QuantizedFlatten"]

import tensorflow as tf
import keras
import functools
import operator

from ..layers_base import (register_quantize_target, tensor_inputs, register_no_output_quantizer,
                           register_aligned_inputs)
from ..recorders import TensorRecorder
from ...tensors import FixedPoint


@register_quantize_target(keras.layers.Flatten)
@register_no_output_quantizer
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedFlatten(keras.layers.Flatten):
    """A Flatten layer that operates on quantized inputs
    """

    @tensor_inputs([FixedPoint])
    def call(self, inputs):
        if not inputs.per_tensor:
            inputs, shift = inputs.expand(inputs.value_bits)
            if getattr(self, 'input_shift', None) is None:
                with tf.init_scope():
                    self.input_shift = TensorRecorder(name=self.name + "/input_shift")
            self.input_shift(shift)
        if tf.executing_eagerly():
            flattened_shape = tf.constant([inputs.shape[0], -1])
        else:
            non_batch_dims = inputs.shape[1:]
            last_dim = int(functools.reduce(operator.mul, non_batch_dims))
            flattened_shape = tf.constant([-1, last_dim])
        return tf.reshape(inputs, flattened_shape)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf
from typing import List

from ...debugging import assert_equal, assert_less_equal, assert_enabled
from .. import FixedPoint, QFloat


@tf.experimental.dispatch_for_api(tf.shape)
def fp_shape(input: FixedPoint, out_type=tf.dtypes.int32, name=None):
    return tf.shape(input.values, out_type, name)


@tf.experimental.dispatch_for_api(tf.shape)
def qf_shape(input: QFloat, out_type=tf.dtypes.int32, name=None):
    return tf.shape(input.fp, out_type, name)


@tf.experimental.dispatch_for_api(tf.reshape)
def fp_reshape(tensor: FixedPoint, shape, name=None):
    if not tensor.per_tensor:
        # Different fractional-bits are defined for the last axis, so
        # it must be preserved during the reshape
        assert_equal(shape[-1], tf.shape(tensor)[-1])
    output = tf.reshape(tensor.values, shape, name)
    return FixedPoint(output, tensor.value_bits, tensor.frac_bits)


@tf.experimental.dispatch_for_api(tf.reshape)
def qf_reshape(tensor: QFloat, shape, name=None):
    if not tensor.per_tensor:
        # Different scales are defined for the last axis, so it must be preserved
        # during the reshape
        assert_equal(shape[-1], tf.shape(tensor.scales)[-1])
    # Reshape inner FixedPoint
    output_fp = tf.reshape(tensor.fp, shape, name)
    return QFloat(output_fp, tensor.scales)


@tf.experimental.dispatch_for_api(tf.transpose)
def fp_transpose(a: FixedPoint, perm=None, conjugate=False, name="transpose"):
    a.assert_per_tensor()
    output = tf.transpose(a.values, perm, conjugate, name)
    return FixedPoint(output, a.value_bits, a.frac_bits)


@tf.experimental.dispatch_for_api(tf.transpose)
def qf_transpose(a: QFloat, perm=None, conjugate=False, name="transpose"):
    a.assert_per_tensor()
    output_fp = tf.transpose(a.fp, perm, conjugate, name)
    return QFloat(output_fp, a.scales)


@tf.experimental.dispatch_for_api(tf.broadcast_to)
def fp_brodcast_to(input: FixedPoint, shape, name=None):
    """Broadcast a FixedPoint tensor for a compatible shape.

    Args:
        input (:obj:`FixedPoint`): a FixedPoint to broadcast.
        shape (tf.Tensor): an 1-D `int` Tensor representing
            the shape of the desired output. Must be one of the
            following types: `int32`, `int64`.
        name (str, optional): a name for the operation. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the brodcasted output. Has the same
            type as `input`.
  """
    # Check first that the last dimension is unchanged
    assert_equal(input.shape[-1], shape[-1], message="To brodcast FixedPoint input,\
                        last dimension should remain unchanged")
    output = tf.broadcast_to(input.values, shape, name)
    return FixedPoint(output, input.value_bits, input.frac_bits)


@tf.experimental.dispatch_for_api(tf.broadcast_to)
def qf_brodcast_to(input: QFloat, shape, name=None):
    """Broadcast a QFloat tensor for a compatible shape.

    Args:
        input (:obj:`QFloat`): a QFloat to broadcast.
        shape (tf.Tensor): an 1-D `int` Tensor representing
            the shape of the desired output. Must be one of the
            following types: `int32`, `int64`.
        name (str, optional): a name for the operation. Defaults to None.

    Returns:
        :obj:`QFloat`: the brodcasted output. Has the same
            type as `input`.
  """
    output_fp = tf.broadcast_to(input.fp, shape, name)
    return QFloat(output_fp, input.scales)


@tf.experimental.dispatch_for_api(tf.concat)
def fp_concat(values: List[FixedPoint], axis, name="concat"):
    """Concatenates FixedPoint tensors along one dimension.

    Args:
        values (List of :obj:`FixedPoint`): List of FixedPoint tensors
            to concatenate.
        axis (list): Dimension along which to concatenate.
        name (str, optional): the name for the Tensorflow ops.
            Defaults to "concat".

    Returns:
        :obj:`FixedPoint`: the concatenate output FixedPoint.
    """
    # For now we only support concatenation of one or two elements
    assert_less_equal(
        len(values),
        2, f"We only support concatenation of one or two FixedPoint. \
           Receives {len(values)} tensors as input.")

    if len(values) == 1:
        return FixedPoint(values[0].values, values[0].value_bits, values[0].frac_bits)

    # Promote to the higher value bits
    if values[0].value_bits < values[1].value_bits:
        values[0] = values[0].promote(values[1].value_bits)
    elif values[0].value_bits > values[1].value_bits:
        values[1] = values[1].promote(values[0].value_bits)

    # For now we don't support concatenation over last dimension
    rank_input = values[0].values.shape.ndims
    if axis < 0:
        dim = axis + rank_input
    else:
        dim = axis

    last_dim = dim == rank_input - 1

    # When dim is not last dimension FixedPoint tensors to concatenate should have the same
    # fractional bits
    if not last_dim:
        assert_equal(values[0].frac_bits, values[1].frac_bits, message=f"The two\
                FixedPoint must have the same frac_bits. Receives {values[0].frac_bits}\
                and {values[1].frac_bits}")
        frac_bits_out = values[0].frac_bits
    else:
        # When dim is last dimension, if one of the two FixedPoint tensors to concatenate is
        # per-tensor, its frac_bits shape is extended like if it was per-axis, then frac_bits can be
        # concatenated
        frac_bits_a = values[0].frac_bits
        frac_bits_b = values[1].frac_bits
        if values[0].per_tensor:
            frac_bits_a = tf.fill((values[0].shape[axis],), values[0].frac_bits)
        if values[1].per_tensor:
            frac_bits_b = tf.fill((values[1].shape[axis],), values[1].frac_bits)
        # Concatenate frac_bits
        frac_bits_out = tf.concat([frac_bits_a, frac_bits_b], axis, f"{name}_frac_bits")

    values_out = tf.concat([values[0].values, values[1].values], axis, name)
    return FixedPoint(values_out, values[0].value_bits, frac_bits_out)


@tf.experimental.dispatch_for_api(tf.concat)
def qf_concat(values: List[QFloat], axis, name="concat"):
    """Concatenates QFloat tensors along one dimension.

    Args:
        values (List of :obj:`QFloat`): List of QFloat tensors
            to concatenate.
        axis (list): Dimension along which to concatenate.
        name (str, optional): the name for the Tensorflow ops.
            Defaults to "concat".

    Returns:
        :obj:`QFloat`: the concatenated output QFloat.
    """
    if len(values) == 1:
        return QFloat(values[0].fp, values[0].scales)

    # For now we only support concatenation of one or two elements
    assert_equal(
        len(values),
        2, f"We only support concatenation of one or two QFloat. \
           Receives {len(values)} tensors as input.")

    # Convenience variables
    a, b = values

    # First, concatenate inner FixedPoint
    output_fp = tf.concat([a.fp, b.fp], axis, name)

    # Evaluate the index of the dimension on which we concatenate
    rank_input = a.values.shape.ndims
    if axis < 0:
        concat_dim = axis + rank_input
    else:
        concat_dim = axis

    if concat_dim == rank_input - 1:
        # When concatenating along the last dimension, we concatenate shapes.
        # If any one of the two QFloat tensors has a scalar shape, its scales
        # shape is extended to a vector as if it was per-axis.
        scales_a = a.scales
        if scales_a.shape.ndims == 0:
            n_scales = tf.shape(a.fp)[-1]
            scales_a = tf.fill((n_scales,), scales_a)
        scales_b = b.scales
        if scales_b.shape.ndims == 0:
            n_scales = tf.shape(b.fp)[-1]
            scales_b = tf.fill((n_scales,), scales_b)
        # The output scales is the concatenation of a and b scales
        output_scales = tf.concat([scales_a, scales_b], axis=-1, name=f"{name}_scales")
    else:
        # When concatenating on another axis, inputs must have the same scales
        assert_equal(values[0].scales, values[1].scales,
                     "We only support concatenation of QFloat inputs with same scales."
                     f"Receives {a.scales} and {b.scales}.")
        output_scales = a.scales

    return QFloat(output_fp, output_scales)


@tf.experimental.dispatch_for_api(tf.stack)
def fp_stack(values: List[FixedPoint], axis, name="stack"):
    """Stack FixedPoint tensors along one dimension.

    Args:
        values (List of :obj:`FixedPoint`): List of FixedPoint tensors
            to stack.
        axis (list): Dimension along which to stack.
        name (str, optional): the name for the Tensorflow ops.
            Defaults to "stack".

    Returns:
        :obj:`FixedPoint`: the stacked output FixedPoint.
    """
    # For now we only support stack of one or two elements
    assert_equal(
        len(values),
        2, f"We only support stack of two FixedPoint. \
           Receives {len(values)} tensors as input.")

    # Promote to the higher value bits
    if values[0].value_bits < values[1].value_bits:
        values[0] = values[0].promote(values[1].value_bits)
    elif values[0].value_bits > values[1].value_bits:
        values[1] = values[1].promote(values[0].value_bits)

    # infer unsigned axis
    offset = values[0].shape.ndims + 1
    if axis < 0:
        dim = axis + offset
    else:
        dim = axis

    last_dim = dim == offset - 1

    # First stack the FixedPoint.values
    values_out = tf.stack([values[0].values, values[1].values], axis, name)
    # When dim is last dimension, the two FixedPoint should be per-tensor otherwise their
    # frac_bits should be equal.
    if last_dim:
        if assert_enabled():
            assert (values[0].per_tensor and values[1].per_tensor), (f"The two \
                    FixedPoint must be per-tensor when stacking on the last dim.\
                    Receives: a.per-tensor:{values[0].per_tensor} and \
                    b.per-tensor:{values[1].per_tensor}")
        frac_bits_a = values[0].frac_bits
        frac_bits_b = values[1].frac_bits
        # Concatenate frac_bits
        frac_bits_out = tf.stack([frac_bits_a, frac_bits_b], 0, f"{name}_frac_bits")
    else:
        assert_equal(values[0].frac_bits, values[1].frac_bits, message=f"The two\
                FixedPoint must have the same frac_bits. Receives {values[0].frac_bits}\
                and {values[1].frac_bits}")
        frac_bits_out = values[0].frac_bits
    return FixedPoint(values_out, values[0].value_bits, frac_bits_out)


@tf.experimental.dispatch_for_api(tf.expand_dims)
def fp_expand_dims(input: FixedPoint, axis, name=None):
    """Returns a tensor with a length 1 axis inserted at index `axis`.

    Args:
        input (FixedPoint): a `Tensor`.
        axis (int): integer specifying the dimension index at which to expand the shape of `input`.
            Given an input of D dimensions, `axis` must be in range `[-(D+1), D]` (inclusive).
        name (str, optional): name of the output `Tensor`. Defaults to None.

    Returns:
        FixedPoint: a tensor with the same data as `input`, with an additional dimension inserted at
            the index specified by `axis`.
    """
    # Evaluate the index of the dimension on which we expand
    rank_input = input.shape.ndims
    if axis < 0:
        expanded_dim = axis + rank_input + 1
    else:
        expanded_dim = axis
    # Only support per-tensor inputs if the expanded along the last dim
    if expanded_dim == rank_input:
        input.assert_per_tensor()

    # Expand dimension on values
    values = tf.expand_dims(input.values, expanded_dim, name)

    # Return a new FixedPoint
    return FixedPoint(values, input.value_bits, input.frac_bits)


@tf.experimental.dispatch_for_api(tf.expand_dims)
def qf_expand_dims(input: QFloat, axis, name=None):
    """Returns a tensor with a length 1 axis inserted at index `axis`.

    Args:
        input (QFloat): a `Tensor`.
        axis (int): integer specifying the dimension index at which to expand the shape of `input`.
            Given an input of D dimensions, `axis` must be in range `[-(D+1), D]` (inclusive).
        name (str, optional): name of the output `Tensor`. Defaults to None.

    Returns:
        QFloat: a tensor with the same data as `input`, with an additional dimension inserted at
            the index specified by `axis`.
    """
    # Evaluate the index of the dimension on which we expand
    rank_input = input.shape.ndims
    if axis < 0:
        expanded_dim = axis + rank_input + 1
    else:
        expanded_dim = axis
    # Only support per-tensor inputs if the expanded along the last dim
    if expanded_dim == rank_input:
        input.assert_per_tensor()

    # Expand dimension on inner FixedPoint
    expanded_fp = tf.expand_dims(input.fp, axis, name)

    # Return a new QFloat
    return QFloat(expanded_fp, input.scales)


@tf.experimental.dispatch_for_api(tf.compat.v1.gather)
def fp_gather(params: FixedPoint, indices, validate_indices=None,
              name=None, axis=None, batch_dims=0):
    """Gather slices from params along axis.

    Args:
        params (:obj:`FixedPoint`): the input FixedPoint.
        indices (int, list): the indices to gather.
        validate_indices (bool, optional): whether to validate the indices. Defaults to None.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.
        axis (int, optional): the axis to gather along. Defaults to None.
        batch_dims (int, optional): the number of batch dimensions to keep. Defaults to 0.

    Returns:
        :obj:`FixedPoint`: a FixedPoint containing the gathered values.
    """
    # If gather along the last axis and params is per axis, select only the targeted frac bits.
    if not params.per_tensor and axis in (-1, len(params.shape) - 1):
        frac_bits = params.frac_bits[indices]
    else:
        frac_bits = params.frac_bits
    # Create a new QTensor, gather the indices in a desired axis
    x_gather = tf.gather(params.values, indices, validate_indices, axis, batch_dims, name)
    return FixedPoint(x_gather, params.value_bits, frac_bits)


@tf.experimental.dispatch_for_api(tf.compat.v1.gather)
def qf_gather(params: QFloat, indices, validate_indices=None, name=None, axis=None, batch_dims=0):
    """Gather slices from params along axis.

    Args:
        params (:obj:`QFloat`): the input QFloat.
        indices (int, list): the indices to gather.
        validate_indices (bool, optional): whether to validate the indices. Defaults to None.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.
        axis (int, optional): the axis to gather along. Defaults to None.
        batch_dims (int, optional): the number of batch dimensions to keep. Defaults to 0.

    Returns:
        :obj:`QFloat`: a QFloat containing the gathered values.
    """
    # If gather along the last axis and params is per axis, select only the targeted frac bits
    # and scales.
    if not params.per_tensor and axis in (-1, len(params.shape) - 1):
        scales = params.scales[indices]
    else:
        scales = params.scales
    # Create a new QTensor, gather the indices in a desired axis

    xfp_gather = tf.gather(params.fp, indices, validate_indices, axis, batch_dims, name)
    return QFloat(xfp_gather, scales)


@tf.experimental.dispatch_for_api(tf.pad)
def fp_pad(tensor: FixedPoint,
           paddings,
           mode="CONSTANT",
           constant_values: FixedPoint = None,
           name=None):
    """Pads a FixedPoint tensor with a mandatory FixedPoint padding value.

    Args:
        tensor (:obj:`FixedPoint`): the input FixedPoint.
        paddings (tf.Tensor): A `Tensor` of type `int32`.
        mode (str): must be "CONSTANT".
        constant_values (:obj:`FixedPoint`): the scalar pad value to use.
        name (str): A name for the operation (optional).

    Returns:
        :obj:`FixedPoint`: a new FixedPoint with padding.
    """
    if mode != 'CONSTANT':
        raise ValueError("When padding a FixedPoint, a constant padding value must be specified.")
    # If the FixedPoint is per-axis, we could pad on any dimension except the last one
    # but we only support per-tensor FixedPoint to limit cognitive and test costs.
    tensor.assert_per_tensor()
    assert_equal(tensor.frac_bits, constant_values.frac_bits,
                 "Padding value is not aligned with the tensor.")
    padded_values = tf.pad(tensor.values, paddings, mode, constant_values.values, name)
    return FixedPoint(padded_values, tensor.value_bits, tensor.frac_bits)


@tf.experimental.dispatch_for_api(tf.pad)
def qf_pad(tensor: QFloat,
           paddings,
           mode="CONSTANT",
           constant_values: QFloat = None,
           name=None):
    """Pads a QFloat tensor with a mandatory QFloat padding value.

    Args:
        tensor (:obj:`QFloat`): the input QFloat.
        paddings (tf.Tensor): A `Tensor` of type `int32`.
        mode (str): must be "CONSTANT".
        constant_values (:obj:`QFloat`): the scalar pad value to use.
        name (str): A name for the operation (optional).

    Returns:
        :obj:`QFloat`: a new QFloat with padding.
    """
    if mode != 'CONSTANT':
        raise ValueError("When padding a QFloat, a constant padding value must be specified.")
    # If the QFloat is per-axis, we could pad on any dimension except the last one
    # but we only support per-tensor QFloat to limit cognitive and test costs.
    tensor.assert_per_tensor()
    assert_equal(tensor.scales, constant_values.scales,
                 "Padding value is not aligned with the tensor.")
    padded_fp = tf.pad(tensor.fp, paddings, mode, constant_values.fp, name)
    return QFloat(padded_fp, tensor.scales)


@tf.experimental.dispatch_for_api(tf.tile)
def fp_tile(input: FixedPoint, multiples, name=None):
    """Constructs a tensor by tiling a given tensor.

    Args:
        input (FixedPoint): a `Tensor`.
        multiples (tf.Tensor): Must be one of the following types: `int32`, `int64`.
            1-D. Length must be the same as the number of dimensions in `input`
        name (str, optional): name of the output `Tensor`. Defaults to None.

    Returns:
        FixedPoint: a tensor with the same data as `input`.
    """
    # If the duplication should occurs on the last axis the input should be per tensor
    if multiples[-1] != 1:
        input.assert_per_tensor()

    # tile input.values "multiples" time
    values = tf.tile(input.values, multiples, name)

    # Return a new FixedPoint
    return FixedPoint(values, input.value_bits, input.frac_bits)


@tf.experimental.dispatch_for_api(tf.tile)
def qf_tile(input: QFloat, multiples, name=None):
    """Constructs a tensor by tiling a given tensor.

    Args:
        input (QFloat): a `Tensor`.
        multiples (tf.Tensor): Must be one of the following types: `int32`, `int64`.
            1-D. Length must be the same as the number of dimensions in `input`
        name (str, optional): name of the output `Tensor`. Defaults to None.


    Returns:
        QFloat: a tensor with the same data as `input`.
    """

    # If the duplication should occurs on the last axis the input should be per tensor
    if multiples[-1] != 1:
        input.assert_per_tensor()

    # tile input.fp "multiples" time
    values = tf.tile(input.fp, multiples, name)

    # Return a new QFloat
    return QFloat(values, input.scales)


@tf.experimental.dispatch_for_api(tf.zeros_like)
def fp_zeros_like(input: FixedPoint, dtype=None, name=None, layout=None):
    """Creates a FixedPoint with all elements set to zero and the same frac_bits and shape as
    input.

    Args:
        input (FixedPoint): a `FixedPoint`.
        dtype (tf.dtype): A type for the returned `Tensor`. Must be `float16`, `float32`,
            `float64`, `int8`, `uint8`, `int16`, `uint16`, `int32`, `int64`, `complex64`,
            `complex128`, `bool` or `string` (optional).
        name (str, optional): name of the output `Tensor`. Defaults to None.
        layout (`tf.experimental.dtensor.Layout`, optional): If provided, the result is a
            [DTensor](https://www.tensorflow.org/guide/dtensor_overview) with the provided layout.
            Defaults to None.

    Returns:
        FixedPoint: a tensor with the same data as `input`.
    """
    # Create a tf.Tensor with all element set to zeros
    values = tf.zeros_like(input.values, dtype, name, layout)

    # Return a new FixedPoint
    return FixedPoint(values, input.value_bits, input.frac_bits)


@tf.experimental.dispatch_for_api(tf.zeros_like)
def qf_zeros_like(input: QFloat, dtype=None, name=None, layout=None):
    """Creates a QFloat with all elements set to zero and the same scales and shape as
    input.

    Args:
        input (QFloat): a `QFloat`.
        dtype (tf.dtype): A type for the returned `Tensor`. Must be `float16`, `float32`,
            `float64`, `int8`, `uint8`, `int16`, `uint16`, `int32`, `int64`, `complex64`,
            `complex128`, `bool` or `string` (optional).
        name (str, optional): name of the output `Tensor`. Defaults to None.
        layout (`tf.experimental.dtensor.Layout`, optional): If provided, the result is a
            [DTensor](https://www.tensorflow.org/guide/dtensor_overview) with the provided layout.
            Defaults to None.

    Returns:
        QFloat: a tensor with the same data as `input`.
    """
    # Create a tf.Tensor with all element set to zeros
    values = tf.zeros_like(input.fp, dtype, name, layout)

    # Return a new QFloat
    return QFloat(values, input.scales)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["measure_layer_quantization_error", "measure_cumulative_quantization_error"]

import keras
import onnx

from .quantization_error import (keras_layer_quantization_error, onnx_node_quantization_error,
                                 keras_cumulative_quantization_error,
                                 onnx_cumulative_quantization_error)


def measure_layer_quantization_error(fmodel, qmodel, target_layer=None, batch_size=16, seed=None):
    """Measures the layer quantization error

    Returns a dictionary where the keys are the name of each layer and the values are a dictionary
    composed of the set of the following metrics:

        * Symmetrical Mean Absolute Percentage Error (SMAPE): :func:`tools.metrics.SMAPE`
        * Saturation: Percentage of how many values in the quantized layer saturate

    Example:
        >>> summary = measure_layer_quantization_error(fmodel, qmodel)
        >>> assert isinstance(summary[a_layer_name], dict)
        >>> assert "SMAPE" in summary[a_layer_name]

    Args:
        fmodel (onnx.ModelProto or tf.keras.Model): the float model.
        qmodel (onnx.ModelProto or tf.keras.Model): the quantized version of `fmodel`.
        target_layer (str, optional): computation error is performed only in the target layer/node,
            expanding the analysis to each output channel. Defaults to None.
        batch_size (int, optional): the batch size of the samples to be generated.
            It allows a better metrics generalization, but consumes more resources. Defaults to 16.
        seed (int, optional): a random seed. Defaults to None.

    Returns:
        dict: the quantization error for each layer

    Notes:
        * Layers/Nodes that do not produce quantization errors will not be taken into account
          (e.g. QuantizedReshape).
    """
    keras_model_types = (keras.Sequential, keras.Model)
    # Check both models have the same type
    if isinstance(fmodel, onnx.ModelProto) and isinstance(qmodel, onnx.ModelProto):
        summary = onnx_node_quantization_error(fmodel,
                                               qmodel,
                                               target_node=target_layer,
                                               batch_size=batch_size,
                                               seed=seed)
    elif isinstance(fmodel, keras_model_types) and isinstance(qmodel, keras_model_types):
        summary = keras_layer_quantization_error(fmodel,
                                                 qmodel,
                                                 target_layer=target_layer,
                                                 batch_size=batch_size,
                                                 seed=seed)
    else:
        model_types = (onnx.ModelProto, *keras_model_types)
        raise TypeError(f"Both models should be the same type, one of {model_types}. "
                        f"Received: {type(fmodel)} and {type(fmodel)}.")
    return summary


def measure_cumulative_quantization_error(fmodel,
                                          qmodel,
                                          target_layer=None,
                                          batch_size=16,
                                          seed=None):
    """Measures the cumulative quantization error

    Returns a dictionary where the keys are the name of each layer and the values are a dictionary
    composed of the set of the following metrics:

        * Symmetrical Mean Absolute Percentage Error (SMAPE): :func:`tools.metrics.SMAPE`
        * Saturation: Percentage of how many values in the quantized layer saturate

    Each metric measures the quantization error from the input to the layer.

    Example:
        >>> summary = measure_cumulative_quantization_error(fmodel, qmodel)
        >>> assert isinstance(summary[a_layer_name], dict)
        >>> assert "SMAPE" in summary[a_layer_name]

    Args:
        fmodel (onnx.ModelProto or tf.keras.Model): the float model.
        qmodel (onnx.ModelProto or tf.keras.Model): the quantized version of `fmodel`.
        target_layer (str, optional): error computation is performed only in the target layer/node,
            expanding the analysis to each output channel. Defaults to None.
        batch_size (int, optional): the batch size of the samples to be generated.
            It allows a better metrics generalization, but consumes more resources. Defaults to 16.
        seed (int, optional): a random seed. Defaults to None.

    Returns:
        dict: the quantization error for each layer

    Notes:
        * Layers/Nodes that do not produce quantization errors will not be taken into account
          (e.g. QuantizedReshape).
    """
    keras_model_types = (keras.Sequential, keras.Model)
    # Check both models have the same type
    if isinstance(fmodel, onnx.ModelProto) and isinstance(qmodel, onnx.ModelProto):
        summary = onnx_cumulative_quantization_error(fmodel,
                                                     qmodel,
                                                     target_node=target_layer,
                                                     batch_size=batch_size,
                                                     seed=seed)
    elif isinstance(fmodel, keras_model_types) and isinstance(qmodel, keras_model_types):
        summary = keras_cumulative_quantization_error(fmodel,
                                                      qmodel,
                                                      target_layer=target_layer,
                                                      batch_size=batch_size,
                                                      seed=seed)
    else:
        model_types = (onnx.ModelProto, *keras_model_types)
        raise TypeError(f"Both models should be the same type, one of {model_types}. "
                        f"Received: {type(fmodel)} and {type(fmodel)}.")
    return summary



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Tools to remove ZeroPadding2D layers from a model.
"""

__all__ = ["remove_zeropadding2d"]

from copy import deepcopy

from keras.models import Sequential
from keras.layers import ZeroPadding2D, Conv2D, SeparableConv2D, DepthwiseConv2D

from .transforms_utils import get_layers, get_layer_index, get_layers_by_type, update_inbound


def _find_removable_zeropadding(model):
    """ Retrieves ZeroPadding2D layers that can be removed.

    This is limited to ZeroPadding2D layers that come before supported layer types and that perform
    a 'same' padding.

    Args:
        model (keras.Model): a model

    Returns:
        dict: map between a ZeroPadding2D and the layer that follows
    """
    map_zeropadding_next = {}

    # Define layers that will support ZeroPadding removal
    supported_layers = (Conv2D, SeparableConv2D, DepthwiseConv2D)

    # Get all ZeroPadding2D layers present in the model
    zeropaddings = get_layers_by_type(model, ZeroPadding2D)

    # Sequential models must be built because a check must be done on shapes
    if zeropaddings and isinstance(model, Sequential) and not model.built:
        raise ValueError("This model has not yet been built.")

    # Find the ones that can be removed
    for zeropadding in zeropaddings:
        # Limit support to single inbound/outbound
        outbounds = zeropadding.outbound_nodes
        if len(zeropadding.inbound_nodes) != 1 or len(outbounds) != 1:
            continue

        # Check that the layer that follows is supported and has a 'valid' padding
        following_layer = outbounds[0].layer
        if not isinstance(following_layer, supported_layers) or following_layer.padding != 'valid':
            continue

        # Check that the combination of ZeroPadding2D + following layer performs a 'same' padding:
        # this is done by checking that next_layer.output_shape * strides = zeropadding.input_shape
        out_spatial_dims = following_layer.output_shape[1:3]
        stride = following_layer.strides
        rectified_out_spatial_dims = tuple(dim * s for dim, s in zip(out_spatial_dims, stride))
        if rectified_out_spatial_dims != zeropadding.input_shape[1:3]:
            continue

        # At this point the ZeroPadding2D is a valid candidate
        map_zeropadding_next[zeropadding] = following_layer
    return map_zeropadding_next


def _get_zeropadding_less_model(model, map_zeropadding_next):
    """ Edits the model configuration to remove ZeroPadding2D layers and rebuilds a model.

    Args:
        model (keras.Model): a model
        map_zeropadding_next (dict): map between a ZeroPadding2D and the layer that follows

    Returns:
        keras.Model: an updated model without ZeroPadding2D layers
    """
    # get_config documentation mentions that a copy should be made when planning to modify the
    # config
    config = deepcopy(model.get_config())
    layers = config['layers']

    for zeropadding, next_layer in map_zeropadding_next.items():
        # Set padding='same' in the layer that follows a ZeroPadding that will be removed
        next_index = get_layer_index(layers, next_layer.name)
        layers[next_index]['config']['padding'] = 'same'

        # For sequential model, the changes stop here: the ZeroPadding2D layers will simply be
        # removed in the following step. For other models, the layers inbounds/outbounds must be
        # rebuilt.
        if isinstance(model, Sequential):
            continue

        # Retrieve the ZeroPadding2D input layer, assuming it has only 1 inbound
        zeropadding_index = get_layer_index(layers, zeropadding.name)
        # tfmot code: 'inbound_nodes' is a nested list where first element is the inbound layername,
        # e.g: [[['conv1', 0, 0, {} ]]]
        updated_inbound = layers[zeropadding_index]['inbound_nodes'][0][0][0]

        # Update ZeroPadding2D outbounds layers: their current inbound is the ZeroPadding2D layer
        # that will be removed so it must be replaced with the ZeroPadding2D previous layer. This
        # results in by-passing the ZeroPadding2D layer: inbound > ZeroPadding2D > outbounds becomes
        # inbound > outbounds.
        update_inbound(layers[next_index], zeropadding.name, updated_inbound)

    # Remove ZeroPadding2D layers
    layers_to_remove = get_layers(config, [zp.name for zp in map_zeropadding_next.keys()])
    for layer_to_remove in layers_to_remove:
        layers.remove(layer_to_remove)

    # Reconstruct model from the config, using the cloned layers
    return model.from_config(config)


def remove_zeropadding2d(model):
    """ Removes ZeroPadding2D layers from a model.

    ZeroPadding2D layers will not be supported by quantization so this transform adds support so
    that when the ZeroPadding2D layers are immediately followed by a convolution layer with 'valid'
    padding, they are removed and the following convolution is updated with a 'same' padding
    instead. This can however only happen when the padding specified in ZeroPadding2D actually
    corresponds to a 'same' padding.

    Args:
        model (keras.Model): the model to update

    Returns:
        keras.Model: the original model or a new model with ZeroPadding2D removed
    """
    # Find ZeroPadding2D and following layer pairs that are candidates for removal
    map_zeropadding_next = _find_removable_zeropadding(model)

    # When there are no valid candidates, return the original model
    if not map_zeropadding_next:
        return model

    # Rebuild a model without ZeroPadding2D by editing the configuration
    updated_model = _get_zeropadding_less_model(model, map_zeropadding_next)

    # Restore model weights
    updated_model.set_weights(model.get_weights())
    return updated_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["QuantizedMaxPool2D", "QuantizedGlobalAveragePooling2D"]

import tensorflow as tf
import keras

from keras.layers import MaxPool2D, GlobalAveragePooling2D

from .layers_base import (register_quantize_target, register_no_output_quantizer, rescale_outputs,
                          tensor_inputs, apply_buffer_bitwidth, QuantizedLayer)
from .quantizers import OutputQuantizer
from ..tensors import FixedPoint, QTensor, QFloat


@register_quantize_target(MaxPool2D)
@register_no_output_quantizer
@keras.saving.register_keras_serializable()
class QuantizedMaxPool2D(QuantizedLayer, MaxPool2D):
    """A max pooling layer that operates on quantized inputs.

    """
    @tensor_inputs([QTensor])
    def call(self, inputs):
        if self.data_format == "channels_last":
            ksize = (1,) + self.pool_size + (1,)
            strides = (1,) + self.strides + (1,)
            data_format = "NHWC"
        else:
            ksize = (1, 1) + self.pool_size
            strides = (1, 1) + self.strides
            data_format = "NCHW"

        padding = self.padding.upper()
        outputs = tf.nn.max_pool(inputs, ksize=ksize, strides=strides, padding=padding,
                                 data_format=data_format)
        return outputs

    def get_config(self):
        return MaxPool2D.get_config(self)


@register_quantize_target(GlobalAveragePooling2D)
@keras.saving.register_keras_serializable()
class QuantizedGlobalAveragePooling2D(QuantizedLayer, GlobalAveragePooling2D):
    """A global average pooling layer that operates on quantized inputs.

     Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """

    def __init__(self, quant_config=None, **kwargs):
        super().__init__(quant_config=quant_config, **kwargs)
        out_quant_cfg = self.quant_config.get("output_quantizer", False)
        if out_quant_cfg:
            self.out_quantizer = OutputQuantizer(
                name="output_quantizer", **out_quant_cfg)
        else:
            self.out_quantizer = None
        self.buffer_bitwidth = apply_buffer_bitwidth(self.quant_config, signed=False)

    def build(self, input_shape):
        super().build(input_shape)
        # Build the spatial size and its reciprocal
        self.spatial_size = (input_shape[1] * input_shape[2])
        self.spatial_size_rec = 1. / self.spatial_size

    @tensor_inputs([QTensor])
    @rescale_outputs
    def call(self, inputs):
        # The only use case where GAP would receive a FixedPoint is when inputs are coming from an
        # add layer and in that case they would necessarily be per-tensor.
        if isinstance(inputs, FixedPoint):
            inputs.assert_per_tensor()
        inputs_sum = tf.reduce_sum(inputs, axis=[1, 2], keepdims=self.keepdims)
        if isinstance(inputs, FixedPoint):
            return QFloat(inputs_sum, self.spatial_size_rec)
        return inputs_sum / self.spatial_size



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Rescaling transformation for quantized models.
"""

__all__ = ["align_rescaling"]

import numpy as np

from copy import deepcopy
from keras.layers import Rescaling, Conv2D, Dense
from keras.saving import serialize_keras_object

from .transforms_utils import get_layers_by_type, get_layers
from ..utils import apply_weights_to_model
from ...layers.convolution import PaddedConv2D


def _find_rescaling_fold_target(rescaling):
    """ Find the folding target and check limitations.

    Args:
        rescaling (keras.layers.Layer): the rescaling layer

    Returns:
        keras.layers.Layer: the layer that follows the Rescaling if a valid candidate, None
        otherwise.
    """
    # Define layers that can accept Rescaling alignment
    supported_dst_layers = [Conv2D, Dense]

    scale_per_axis = isinstance(rescaling.scale, (list, tuple)) and len(rescaling.scale) > 1
    if isinstance(rescaling.offset, (list, tuple)):
        zero_offset = all(offset == 0 for offset in rescaling.offset)
    else:
        zero_offset = rescaling.offset == 0
    if not scale_per_axis and zero_offset:
        # Rescaling is already aligned: nothing to do
        return None
    # Alignment is limited to single outbound node Rescaling layers
    if len(rescaling.outbound_nodes) != 1:
        raise ValueError("Found a non-aligned Rescaling layer in the model with multiple outbounds "
                         "which is not supported.")
    # Retrieve the destination layer and check its type
    dst_layer = rescaling.outbound_nodes[0].layer
    if type(dst_layer) not in supported_dst_layers:
        raise ValueError(f"Layer type {type(dst_layer)} after Rescaling not supported, must be in "
                         f"{supported_dst_layers}.")
    return dst_layer


def _adapt_padding(model, offset, dst_layer):
    """ Update padding scheme by replacing Conv2D with a PaddedConv2D and appropriate padding value.

    Args:
        model (keras.Model): the original model
        offset (float, list, tuple): the offset to fold
        dst_layer (keras.layer): the layer where offset or scale will be folded

    Returns:
        keras.Model: the updated model
    """
    # Copy configuration before applying modifications
    config = deepcopy(model.get_config())
    # Fold rescaling by editing the model configuration
    dst_config = get_layers(config, [dst_layer.name])[0]
    # Force bias
    dst_config['config']['use_bias'] = True
    # Replace Conv2D with 'same' padding by PaddedConv2D with correct padding value
    if isinstance(dst_layer, Conv2D) and dst_layer.padding.lower() == 'same':
        if isinstance(offset, (list, tuple)):
            chan = dst_layer.get_weights()[0].shape[2]
            assert len(offset) in (1, chan), "offset must be a scalar or of size of channels"
            pad_values = [float(-p) for p in offset]
        else:
            pad_values = float(-offset)

        new_config = PaddedConv2D.from_config(dst_config['config'])
        dst_config.update(serialize_keras_object(new_config))
        dst_config['config']['padding_value'] = pad_values

    # Reconstruct model from the config
    aligned_model = model.from_config(config)

    # Restore model weights
    variables_dict = {var.name: var for var in model.variables}
    apply_weights_to_model(aligned_model, variables_dict, False)
    return aligned_model


def _fold_rescaling(rescaling_layer, dst_layer, had_bias):
    """ Fold rescaling parameters into dst_layer weights and bias.

    Note that if scales are per-channel, there are folded in the weights.

    Args:
        rescaling_layer (keras.layers.Layer): the Rescaling layer
        dst_layer (keras.layers.Layer): the layer where Rescaling is folded
        had_bias (bool): whether the original layer had a bias or not
    """
    assert isinstance(dst_layer, (Conv2D, Dense))
    base_weights = dst_layer.get_weights()
    new_w = base_weights[0].copy()
    filters = new_w.shape[-1]

    scale = rescaling_layer.scale
    if isinstance(scale, (list, tuple)) and len(scale) > 1:
        # If scale is not a scalar, align it
        rescaling_layer.scale = 1
        for i in range(filters):
            # To compensate, adjust the weights of the next layer
            new_w[..., i] *= scale
        if isinstance(dst_layer, PaddedConv2D):
            # Also rescale the padding value
            padding_value = np.array(dst_layer._padding_value, dtype=np.float32)
            dst_layer._padding_value = list(padding_value / scale)
    new_weights = [new_w]

    if dst_layer.use_bias:
        # Build zero initialized biases if the original layer didn't have any
        new_biases = base_weights[1].copy() if had_bias else np.zeros(filters)
        for i in range(filters):
            # Rescale biases filter by filter to enable broadcast if offsets are per channel
            w_i = base_weights[0][..., i]
            new_biases[i] += np.sum(w_i * rescaling_layer.offset)
        new_weights += [new_biases]
        rescaling_layer.offset = 0
    dst_layer.set_weights(new_weights)


def align_rescaling(model):
    """Aligns the Rescaling layer of the model to make it quantization ready.

    This folds the offset into the bias of next layer.

    The resulting Rescaling is therefore compatible with a quantization to a
    QuantizedRescaling.

    If the source model does not contain a Rescaling or if its Rescaling is already
    aligned, then the original model is returned.

    Args:
        model (keras.Model): the source Keras model

    Returns:
        keras.Model: the original model or a new model with Rescaling layer aligned
    """
    # Check if the model has a Rescaling layer and return the original model if not
    rescaling_layer = get_layers_by_type(model, Rescaling)
    if not rescaling_layer:
        return model

    # Limit alignment to the first rescaling layer (a model should only have one)
    rescaling_layer = rescaling_layer[0]

    # Find folding target and check limitations
    dst_layer = _find_rescaling_fold_target(rescaling_layer)

    # If no folding target was found return the original model
    if dst_layer is None:
        return model

    # There is a rescaling offset, dst_layer padding scheme must be updated
    offset = rescaling_layer.offset
    aligned_model = _adapt_padding(model, offset, dst_layer)

    # Fold Rescaling parameters into the new layer weights
    _fold_rescaling(aligned_model.get_layer(rescaling_layer.name),
                    aligned_model.get_layer(dst_layer.name),
                    dst_layer.use_bias)
    return aligned_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["insert_rescaling"]

import numpy as np
import onnx.numpy_helper
from onnx.helper import make_node

from ...graph_tools import get_field
from ...layers import DOMAIN
from ..model import ONNXModel


def _add_rescale_node(onnx_model, input_name, output_name, scale=1.0, offset=0.0, perm=[]):
    """Creates a custom Rescale node as the sequence of (Tranpose) -> Mul -> (Add) operations

    Args:
        onnx_model (ONNXModel): The ONNX model to which the rescale node will be added.
        input_name (str): The name of the input tensor for the rescale operation.
        output_name (str): The name of the output tensor for the rescale operation.
        scale (float or list, optional): The scaling factors to apply to the model inputs.
            Defaults to 1.0.
        offset (float or list, optional): The offset values to apply after scaling the model
            inputs. Defaults to 0.0.
        perm (list, optional): The permutation to apply to the dimensions of the rescale
            node inputs. Defaults to [].
    """
    def _format_weight(x):
        x = np.array(x, dtype="float32")
        if needs_to_tranpose and x.size != 1:
            # Expand dims to support permutation
            x = np.expand_dims(x, axis=tuple(range(len(perm) - x.ndim)))
            x = np.transpose(x, perm)
        return x

    nodes, weights = [], []
    inode = input_name

    # Create a Transpose node if permutation changes the order of inputs
    needs_to_tranpose = any(x != idx for idx, x in enumerate(perm))
    if needs_to_tranpose:
        nodes.append(make_node('Transpose',
                               inputs=[inode],
                               outputs=[f"{input_name}/transposed"],
                               perm=perm))
        inode = nodes[-1].output[0]

    # Create a Scale node
    # Note we need to permute scale since transpose is applied as first operation
    nodes.append(make_node("Mul",
                           inputs=[inode, f"{input_name}/input_scale"],
                           outputs=[f"{input_name}/scaled"]))
    weights.append(onnx.numpy_helper.from_array(_format_weight(scale), nodes[-1].input[1]))
    inode = nodes[-1].output[0]

    # Create an Offset node if needed
    # Note we need to permute scale since transpose is applied as first operation
    if np.any(offset != 0.0):
        nodes.append(make_node("Add",
                               inputs=[inode, f"{input_name}/input_offset"],
                               outputs=[f"{input_name}/shifted"]))
        weights.append(onnx.numpy_helper.from_array(_format_weight(offset), nodes[-1].input[1]))

    # Replace last name if there are at least one node to append
    if len(nodes) > 0:
        nodes[-1].output[0] = output_name

    # Add nodes to onnx/weights to model
    onnx_model.initializer_extend(weights)
    onnx_model.add_nodes(nodes)


def insert_rescaling(model):
    """Insert a Custom Rescaling node in the model which applies a scaling factor,
    offset, and optional transposes the inputs.

    Args:
        model (ONNXModel): The ONNX model to be processed.
    """
    def _get_variables(node):
        values = []
        for inode in node.input:
            try:
                values.append(model.get_variable(inode))
            except AssertionError:
                values.append(None)
        return values

    assert isinstance(model, ONNXModel)
    assert len(model.input) == 1, "Only a single input is supported"

    # Start with the first node only if the input is connected to one node
    # Otherwise, create a fake node to skip main loop
    first_nodes = model.input_name_to_nodes()[model.input[0].name]
    node = first_nodes[0] if len(first_nodes) == 1 else onnx.NodeProto()

    # We don't add rescaling node if the model is quantized
    if node.op_type in ("InputQuantizer") and node.domain == DOMAIN:
        return

    # Default Rescale parameters
    scale, offset, perm = 1, 0, []
    nodes_to_remove = []

    # Main loop
    while node.op_type in ("Mul", "Add", "Sub"):
        first_value, second_value = _get_variables(node)
        if first_value is None and second_value is None:
            # Nothing to do if there is no initializer
            break

        # Apply scale/offset to current values
        # Note that it does not make sense for both inputs
        # to be initializers (the graph would be disconnected).
        new_value = first_value if first_value is not None else second_value
        if node.op_type == "Mul":
            # Scale and offset are both multiplied by the Mul scale
            scale *= new_value
            offset *= new_value
        elif node.op_type == "Add":
            offset += new_value
        elif node.op_type == "Sub":
            # The operation changes according to the position of the initializer
            if first_value is None:
                # y = (scale*X + offset) - new_value = scale*X + offset - new_value
                offset -= new_value
            else:
                # y = new_value - (scale*X + offset) = -scale*X + new_value - offset
                offset = new_value - offset
                scale *= -1
        nodes_to_remove.append(node)

        # Break loop if current node has multiple outbounds
        outbounds = model.get_children(node)
        if len(outbounds) != 1:
            break

        # Get next node
        node = outbounds[0]

    if node.op_type == "Transpose":
        perm = get_field(node, "perm")
        nodes_to_remove.append(node)

    # If no nodes to remove are found, we create a new value info.
    # This value info will be the output of the rescale node, as the input
    # of the first node is the graph's input and will be redirected to the rescale node.
    if nodes_to_remove == []:
        rescale_output_name = f"{model.input[0].name}/rescaled"
        rescale_output_value_info = model.find_value_info_by_name(
            model.input[0].name).__deepcopy__()
        rescale_output_value_info.name = rescale_output_name
        model.graph().value_info.append(rescale_output_value_info)
        model.replace_input_of_all_nodes(model.input[0].name, rescale_output_name)
    else:
        rescale_output_name = nodes_to_remove[-1].output[0]

    # Add the rescale node as a sequence of (Mul) + (Add) + (Transpose)
    rescale_input_name = model.input[0].name
    _add_rescale_node(model, rescale_input_name, rescale_output_name, scale, offset, perm)
    model.remove_nodes(nodes_to_remove)

    # As we add new nodes, we need to topologically sort the model graph
    model.topological_sort()
    model.clean_initializers()
    model.check_model()



from .sanitize import *
from .remove_pad_nodes import *
from .convert_avg_to_gap import *
from .remove_reshape import *
from .convert_conv_to_gemm import *
from .squeeze_reshape_to_flatten import *
from .convert_matmul_to_gemm import *
from .convert_min_max_to_clip import *
from .fuse_relu_clip_nodes import *
from .untranspose_gemm_weights import *
from .insert_rescaling import *



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["AKIDA_ONNX_LAYERS"]

from ..graph_tools import infer_partial_io

AKIDA_ONNX_LAYERS = []


def infer_function_parameters(nodes):
    """Provides the expected input, output and attribute names for
    the node sequence to define a function based on them.

    Args:
        nodes (list of NodeProto): the node sequence.

    Returns:
        list: the expected input, output and attribute names
    """
    # Infer ordered, not null and unique input/output names
    inputs, outputs = infer_partial_io(nodes)

    # Handle attribute list.
    attributes = set()
    for node in nodes:
        # We consider as attributes those with ref_attr_name != "":
        # it is the real link between node attribute names and attribute in function.
        node_attrs = [attr for attr in node.attribute if attr.ref_attr_name]
        new_attributes = set(attr.ref_attr_name for attr in node_attrs)
        attributes.update(new_attributes)
    return inputs, outputs, list(attributes)


def register_new_subgraph(func):
    """Try to register a new subgraph (view as a function in ONNX) under brainchip domain.

    Args:
        func (FunctionProto): the function to register.
    """
    global AKIDA_ONNX_LAYERS

    # Only register new functions (avoid duplicates)
    if all(f.name != func.name for f in AKIDA_ONNX_LAYERS):
        AKIDA_ONNX_LAYERS.append(func)
    return func



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["remove_reshape"]

from ...graph_tools import get_tensor_shape
from ..model import ONNXModel


def _remove_redudant_flatten_reshape(model):
    def _find_reshape_flatten_sequence(model):
        expected_op_type = ("Reshape", "Flatten")
        for node in model.nodes():
            # Reshape should have only one non initializer, Flatten has by default one
            if len(model.get_node_inputs(node)) > 1:
                continue

            node_outbounds = model.get_children(node)

            # Find candidate node and get the type of the next node if there is one
            if (node.op_type in expected_op_type) and len(node_outbounds) == 1:
                outbound = node_outbounds[0]
                if outbound.op_type in expected_op_type:
                    return node
        return None

    # We consider successively the following sequences of nodes:
    # Reshape > Reshape or Reshape > Flatten or Flatten > Reshape or Flatten > Flatten
    while True:
        # Find a node that matches with the nodes pattern
        node_to_remove = _find_reshape_flatten_sequence(model)
        if node_to_remove is None:
            break  # No more node to delete

        # We remove the selected node
        model.remove_node(node_to_remove, update_graph=True)


def _remove_pointeless_reshape_flatten(model):
    nodes_to_remove = []

    for node in model.nodes():
        if node.op_type in ('Reshape', 'Flatten'):
            if len(model.get_node_inputs(node)) > 1:
                continue

            input_value_info = model.find_value_info_by_name(node.input[0])
            output_value_info = model.find_value_info_by_name(node.output[0])

            input_shape = get_tensor_shape(input_value_info)
            output_shape = get_tensor_shape(output_value_info)
            if input_shape == output_shape:
                nodes_to_remove.append(node)

    # Remove Reshape/Flatten nodes
    model.remove_nodes(nodes_to_remove, update_graph=True)


def remove_reshape(model):
    """Removes pointless Reshape/Flatten nodes (which input shape is the same as output shape)
    from the ONNX model. It also applies Multiple Reshape/Flatten transformations for
    the following patterns:
        - Reshape + Reshape → Reshape
        - Flatten + Flatten → Flatten
        - Reshape + Flatten → Flatten
        - Flatten + Reshape → Reshape

    Args:
        model (ONNXModel): The ONNX model to be processed.
    """
    assert isinstance(model, ONNXModel)

    _remove_redudant_flatten_reshape(model)
    _remove_pointeless_reshape_flatten(model)

    # Clean shape from initializers
    model.clean_initializers()
    model.check_model()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Helper that prepares a model for quantization.
"""

__all__ = ['sanitize']

from . import (align_rescaling, invert_batchnorm_pooling, fold_batchnorms, remove_zeropadding2d,
               invert_relu_maxpool, replace_lambda, convert_conv_to_dw_conv, remove_reshape,
               fold_activations_in_add, split_layers_with_relu)


def sanitize(model):
    """ Sanitize a model preparing it for quantization.

    This is a wrapping successive calls to several model transformations which aims at making the
    model quantization ready.

    Args:
        model (keras.Model): the input model

    Returns:
        keras.Model: the sanitized model
    """

    # Replace lambda layers
    model = replace_lambda(model)

    # Splits layers with 'relu' activation into two separate layers
    # one without activation and one with a separate 'relu' activation layer.
    model = split_layers_with_relu(model)

    # Replace Conv2D layers that behave as DepthwiseConv2D to the latest.
    model = convert_conv_to_dw_conv(model)

    # Multiple Reshape/Flatten removal transformation
    model = remove_reshape(model)

    # Align Rescaling (if needed)
    model = align_rescaling(model)

    # Invert ReLU <-> MaxPool layers so that MaxPool comes first
    model = invert_relu_maxpool(model)

    # Invert BN <-> Pooling layers and fold BN into their preceding layers
    model = invert_batchnorm_pooling(model)
    model = fold_batchnorms(model)

    # Fold ReLUs with no max_value that comes after an Add layer
    model = fold_activations_in_add(model)

    # Remove unsupported ZeroPadding2D layers and replace them with 'same' padding convolution when
    # possible
    model = remove_zeropadding2d(model)
    return model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["QuantizedRescaling"]

import keras.layers
import tensorflow as tf
import keras

from .layers_base import (register_quantize_target, register_no_output_quantizer, tensor_inputs,
                          QuantizedLayer)
from ..tensors import QFloat


@register_quantize_target(keras.layers.Rescaling)
@register_no_output_quantizer
@keras.saving.register_keras_serializable()
class QuantizedRescaling(QuantizedLayer, keras.layers.Rescaling):
    """A layer that multiplies integer inputs by a scale

    This is a simplified version of the keras Rescaling layer:

    - it only supports a scalar scale,
    - it only supports zero offsets.

    This layer assumes the inputs are 8-bit integer: it simply wraps them into
    an 8-bit per-tensor QFloat with the specified scale.

    Args:
        scale (float): a scalar scale.
    """

    def __init__(self, scale, **kwargs):
        super().__init__(scale, **kwargs)
        if tf.rank(self.scale) > 0:
            raise ValueError("QuantizedRescaling only accepts scalar scale.")
        if tf.reduce_any(self.offset != 0):
            raise ValueError("QuantizedRescaling only accepts zero offset.")

    @tensor_inputs([tf.Tensor])
    def call(self, inputs):
        # Wrap them into a QFloat with the specified scale
        return QFloat(inputs, self.scale)

    def get_config(self):
        return keras.layers.Rescaling.get_config(self)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
ReLU > MaxPool inversion helper.
"""

__all__ = ["invert_relu_maxpool"]

from keras.layers import MaxPool2D, ReLU

from .transforms_utils import find_layers_pairs, invert_layer_pairs


def invert_relu_maxpool(model):
    """ Inverts ReLU and MaxPool2D layers in a model to have MaxPool2D before ReLU.

    This transformation produces a strictly equivalent model.

    Args:
        model (keras.Model): a model

    Returns:
        keras.Model: keras.Model: the original model or the updated model
    """
    # Find ReLU followed by MaxPool2D layer pairs that are candidates for inversion
    map_relu_mp = find_layers_pairs(model, ReLU, MaxPool2D)

    # When there are no valid candidates, return the original model
    if not map_relu_mp:
        return model

    # Rebuild a model with MP and ReLU inverted by editing the configuration
    return invert_layer_pairs(model, map_relu_mp)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Layer decorators.
"""

import inspect
import tensorflow as tf
from copy import deepcopy

from .quantization_params import get_quantization_params
from .quantizers import WeightQuantizer, AlignedWeightQuantizer, OutputQuantizer
from .recorders import TensorRecorder
from ..tensors import FixedPoint


class QuantizedLayer(tf.keras.layers.Layer):
    """Interface to define a quantized layer

    It allows to verify if the arguments match with those expected by ``arg_constraints``
    and remove those in ``ignored_args``.

    Both should be defined as class variables.

    Notes:
        * ``ignored_args`` is a list with the argument names to be skipped
        * ``arg_constraints`` is a dictionnary where each set of (key, value) corresponds
          to the argument name and the allowed value. For multiple ones, it may be a callable
          that returns the list of values (e.g. lambda: [value1, value2, ...])
    """

    def __init__(self, *args, quant_config=None, **kwargs):
        check_arg_constraints(self, kwargs)
        self._pop_ignored_args(kwargs)
        self.quant_config = init_quant_config(quant_config)
        super().__init__(*args, **kwargs)

    def _pop_ignored_args(self, config):
        """Function to handle arguments that should be ignored, i.e. dropped from config

        Args:
            config (dict): the argument list
        """
        if not isinstance(ignored_args := getattr(self, 'ignored_args', []), list):
            raise RuntimeError(f"'ignored_args' in {self.name} must be a list.")
        for arg in ignored_args:
            config.pop(arg, None)

    def build(self, input_shape):
        # Ensure variables are build with the appropriate name
        with tf.name_scope(self.name + '/'):
            super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        self._pop_ignored_args(config)
        config.update({"quant_config": deepcopy(self.quant_config)})
        return config


# Mapper to match float layer with its quantized version, populated thanks to
# `register_quantize_target` decorator on quantized layers.
_GLOBAL_LAYER_TO_QLAYER = {}

# List of quantized layers that cannot rescale their outputs (no output quantizer)
_GLOBAL_NO_OUTPUT_QUANTIZER = []

# List of quantized layers that require aligned inputs
_GLOBAL_ALIGNED_INPUTS = []


def register_quantize_target(base_layer):
    """Register the current class as a target for the quantization of a keras layer.

    This decorator injects the decorated class into the _GLOBAL_LAYER_TO_QLAYER dictionary, so that
    it registered as the quantization target for the provided `base_layer`.

    Args:
        base_layer (keras.Layer, list): the origin layer (or list of) that should be quantized as
            the current class

    Returns:
        Callable: a decorator that registers the decorated class
    """
    def _register_target(target, arg):
        """Register the current class as a target for the quantization of a keras layer.

        Args:
            target (keras.Layer): the target to register
            arg (Cls): the current class to register
        """
        base_class_name = target.__name__

        if not inspect.isclass(arg):
            raise ValueError("Can only register class objects with 'register_quantize_target'.")

        if base_class_name in _GLOBAL_LAYER_TO_QLAYER:
            raise ValueError(f"{base_class_name} has already been registered to "
                             f"{_GLOBAL_LAYER_TO_QLAYER[base_class_name]}.")

        _GLOBAL_LAYER_TO_QLAYER.update({base_class_name: arg})

    def decorator(arg):
        targets = base_layer if isinstance(base_layer, list) else [base_layer]
        for target in targets:
            _register_target(target, arg)
        return arg

    return decorator


def register_no_output_quantizer(arg):
    """Register the decorated class as not able to rescale its outputs.

    _GLOBAL_NO_OUTPUT_QUANTIZER is populated with the quantized layer type.

    Args:
        arg (Cls): the class to register

    Returns:
        Callable: a decorator that registers the decorated class
    """
    if not inspect.isclass(arg):
        raise ValueError("Can only register class objects with 'register_no_output_quantizer'.")
    _GLOBAL_NO_OUTPUT_QUANTIZER.append(arg)
    return arg


def register_aligned_inputs(arg):
    """Register the decorated class as requiring aligned inputs.

    _GLOBAL_ALIGNED_INPUTS is populated with the quantized layer type.

    Args:
        arg (Cls): the class to register

    Returns:
        Callable: a decorator that registers the decorated class
    """
    if not inspect.isclass(arg):
        raise ValueError("Can only register class objects with 'register_aligned_inputs'.")
    _GLOBAL_ALIGNED_INPUTS.append(arg)
    return arg


def rescale_outputs(call):
    """ Decorator to rescale the outputs produced by a layer 'call' function.

    Args:
        call (Callable): the decorated call function

    Returns:
        Callable: the decorated function
    """

    def decorator(self, inputs):
        outputs = call(self, inputs)
        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs)
        return outputs
    return decorator


def tensor_inputs(supported):
    """ Decorator to check the input tensors passed to a layer `call`.

    Args:
        supported (list): list of supported input types

    Returns:
        Callable: the decorated function
    """

    def decorator(call):
        if not isinstance(supported, (list, tuple)):
            raise TypeError(f"'supported' must be a list or a tuple, received {type(supported)}.")

        def check_inputs(self, inputs):
            # Raise an error if the inputs are not in the 'supported' types
            if not isinstance(inputs, tuple(supported)):
                raise TypeError(f"{self.__class__.__name__} only accepts {supported} inputs. "
                                f"Receives {type(inputs)} inputs.")

            if isinstance(inputs, tf.Tensor):
                if not inputs.dtype.is_integer:
                    # Remove decimal part, through an intermediate cast
                    inputs = tf.cast(inputs, tf.int32)
                # Default to 8bit but 16bit (16bits -1bit for the sign) possible with TENN
                # sequential models
                in_bits = 15 if inputs.dtype == tf.int16 else 8
                inputs = FixedPoint(inputs, in_bits, 0)
            # If the layer is a (Depthwise)BTC the input alignement is handled directly in the
            # layer call method, after the input is added to the fifo.
            if self.__class__.__name__ in ['QuantizedBufferTempConv',
                                           'QuantizedDepthwiseBufferTempConv']:
                return call(self, inputs)
            if getattr(self, 'buffer_bitwidth', None) is not None:
                if (isinstance(inputs, FixedPoint) and not inputs.per_tensor
                        and self.__class__.__name__ not in
                        ['QuantizedDepthwiseConv2D', 'QuantizedExtractToken', 'QuantizedReLU']):
                    # Expand the inputs to a higher bitwidth to avoid saturation and align them.
                    # Depthwise layers do not require input_shift because each channel is
                    # handled by a single filter.
                    inputs, shift = inputs.expand(self.buffer_bitwidth)
                    if getattr(self, 'input_shift', None) is None:
                        # Add object that will store the shift values.
                        # From Keras documentation, any variable creation taking place in call
                        # should be wrapped with tf.init_scope
                        with tf.init_scope():
                            self.input_shift = TensorRecorder(name=self.name + "/input_shift")
                    self.input_shift(shift)
                else:
                    # Promote inputs to avoid a saturation
                    inputs = inputs.promote(self.buffer_bitwidth)
            return call(self, inputs)
        return check_inputs
    return decorator


def neural_layer_init(separable):
    """ Decorator to initialize a neural layer.

    Args:
        separable (bool): True if the layer has separable weights.

    Returns:
        Callable: the decorated function
    """
    def decorator(init):
        def wrapper(self, *args, **kwargs):
            # First call super().__init__
            super_init = getattr(super(type(self), self), "__init__")
            # Checks params
            activation = kwargs.get('activation')
            if activation and activation != 'linear':
                raise ValueError(f"{self.__class__.__name__} does not support activation. "
                                 f"Receives {activation} activation.")
            # Handle special parameter "padding_value" that must not be passed to super()
            updated_kwargs = kwargs.copy()
            updated_kwargs.pop('padding_value', None)
            super_init(*args, **updated_kwargs)

            # Check data format
            if hasattr(self, "data_format") and self.data_format != "channels_last":
                raise TypeError(f"{self.__class__.__name__} only accepts channels_last data_format."
                                f" Receives {self.data_format}.")
            # Then start neural layer init
            default_weight_bits = get_quantization_params().weight_bits

            # Use quant_config to build quantizers
            if separable:
                # Separable layer has two weights quantizers to handle different max values
                if "dw_weight_quantizer" not in self.quant_config:
                    self.quant_config["dw_weight_quantizer"] = {"bitwidth": default_weight_bits}
                dw_weight_quantizer_cfg = self.quant_config["dw_weight_quantizer"]
                # Separable depthwise weights are quantized per-tensor
                dw_weight_quantizer_cfg.update({'axis': None})
                self.dw_weight_quantizer = WeightQuantizer(name="dw_weight_quantizer",
                                                           **dw_weight_quantizer_cfg)
                if "pw_weight_quantizer" not in self.quant_config:
                    self.quant_config["pw_weight_quantizer"] = {"bitwidth": default_weight_bits}
                pw_weight_quantizer_cfg = self.quant_config["pw_weight_quantizer"]
                self.pw_weight_quantizer = WeightQuantizer(name="pw_weight_quantizer",
                                                           **pw_weight_quantizer_cfg)
            else:
                if "weight_quantizer" not in self.quant_config:
                    self.quant_config["weight_quantizer"] = {"bitwidth": default_weight_bits}
                weight_quantizer_cfg = self.quant_config["weight_quantizer"]
                self.weight_quantizer = WeightQuantizer(name="weight_quantizer",
                                                        **weight_quantizer_cfg)

            # Finalize output and bias quantizers
            out_quant_cfg = self.quant_config.get("output_quantizer", False)
            if out_quant_cfg:
                self.out_quantizer = OutputQuantizer(name="output_quantizer", **out_quant_cfg)
            else:
                self.out_quantizer = None
            if self.use_bias:
                bias_quantizer_cfg = self.quant_config.get("bias_quantizer", {})
                self.bias_quantizer = AlignedWeightQuantizer(name="bias_quantizer",
                                                             **bias_quantizer_cfg)
            self.buffer_bitwidth = apply_buffer_bitwidth(self.quant_config, signed=True)

            # Baseline init
            init(self, *args, **kwargs)
        return wrapper
    return decorator


def init_quant_config(quant_config):
    """Ensures that modifications do not affect the original quantization configuration.

    Args:
        quant_config (dict): the serialized quantization configuration.

    Returns:
        dict: a serialized quantization configuration.
    """
    if isinstance(quant_config, dict):
        return deepcopy(quant_config)
    return dict()


def apply_buffer_bitwidth(quant_config, signed=True):
    """ Read buffer_bitwidth on quant_config. If it is missing,
    set 'buffer_bitwidth' parameter in ``quant_config`` with default value given by the context.

    Args:
        quant_config (dict): the serialized quantization configuration.
        signed (bool, optional): whether operations will be done with sign. Defaults to True.

    Returns:
        int: buffer bitwidth value
    """
    buffer_bits = quant_config.get("buffer_bitwidth", None)
    if buffer_bits is None:
        buffer_bits = get_quantization_params().buffer_bits
        quant_config["buffer_bitwidth"] = buffer_bits
    if signed:
        buffer_bits -= 1
    return buffer_bits


def check_arg_constraints(layer, config):
    """Function to check unsupported arguments in config

    Args:
        layer (keras.Layer): layer to extract unsupported arguments
        config (dict): the list of arguments to check

    Note: This function does not remove arg from config, given there are layers
        that still need these parameters (with their default value). Please use
        pop_ignored_args if they have to be removed.
    """
    if not isinstance(arg_constraints := getattr(layer, 'arg_constraints', {}), dict):
        raise RuntimeError(f"'arg_constraints' in {layer.name} must be a dict.")
    for arg, arg_value in arg_constraints.items():
        supported_values = arg_value() if callable(arg_value) else [arg_value]
        if arg in config and not any(config[arg] == x for x in supported_values):
            lname = getattr(layer, 'name', config.get('name', layer.__class__.__name__))
            raise RuntimeError(
                f"Argument '{arg}' in layer '{lname}' is only supported with "
                f"one of {supported_values}. Receives '{config[arg]}'.")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["OnnxLayer", "DOMAIN", "VERSION"]

import numpy as np
import uuid
from collections import defaultdict
import inspect

from onnx import ValueInfoProto
from onnx.helper import (make_function, make_node, make_opsetid, make_tensor_value_info,
                         np_dtype_to_tensor_dtype)
from onnx.defs import onnx_opset_version

from .register import register_new_subgraph, infer_function_parameters
from ..graph_tools import to_field, value_info_to_tensor_shape, array_to_tp

DOMAIN = "com.brainchip"
VERSION = 1

# List of quantized layers that requires a power of two scale
_GLOBAL_REQUIRES_FP_INPUTS = []

# List of quantized layers that should always output in 8bit
_GLOBAL_REQUIRES_DOWNSCALE = []


def register_node_format(requires_downscale=False, requires_fp_inputs=False):
    """Decorator to register the node format requirements such as whether it requires
    downscaling or fixed-point inputs.

    _GLOBAL_REQUIRES_FP_INPUTS and _GLOBAL_REQUIRES_DOWNSCALE are populated
    with the quantized node type.

    Args:
        requires_downscale (bool, optional): determines if the node class requires downscaling.
            Defaults to False.
        requires_fp_inputs (bool, optional): determines if the node class requires fixed-point
            inputs. Defaults to False.

    Returns:
        Callable: a decorator that registers the decorated class
    """
    def decorator(cls):
        if not inspect.isclass(cls):
            raise ValueError("Can only register class objects with 'register_node_format'.")
        if requires_downscale:
            _GLOBAL_REQUIRES_DOWNSCALE.append(cls)
        if requires_fp_inputs:
            _GLOBAL_REQUIRES_FP_INPUTS.append(cls)
        return cls
    return decorator


class OnnxLayer:
    """Abstract class that represents an onnx subgraph in brainchip domain.

    Child must define the attributes on __init__ and return the node list (subgraph) on
    build_subgraph(). If these requirements are met, make_node() could be used to
    define/register the custom node.

    Args:
        base_name (str): the operation type base name.
        name (str, optional): the node name. Defaults to ''.
        kwargs (dict, optional): the custom attributes. Each attribute type will be
            infered by ``onnx.helper.make_attribute()``. Defaults to {}.
    """

    def __init__(self, base_name, name='', **kwargs):
        self.base_name = base_name
        self.name = name
        self._input = None
        self._output = None
        self.serialize_attr = defaultdict(bool)

        # Load attributes
        # Note: this field is called 'attribute' to align it to the same ONNX standard
        self.attribute = self._load_attributes(**kwargs)

        # Create empty variable to save the weights
        self._weights = {}

    @property
    def op_type(self):
        op_name = self.base_name
        if self.serialize_attr["flatten"]:
            op_name += "Flatten"
        bias = self.weights.get("bias", np.array([]))
        if bias.size > 0:
            op_name += "Biased"
        pool_type = self.serialize_attr["pool_type"]
        if pool_type == "max":
            op_name += "MaxPool"
        elif pool_type == "gap":
            op_name += "GlobalAvgPool"
        if self.serialize_attr["activation"]:
            op_name += "ReLU"
            # We assume unbounded activation when max_value = 0
            max_value = self.weights.get("max_value", np.array([]))
            if np.any(max_value != 0):
                op_name += "Clipped"
        if self.serialize_attr["scale"]:
            op_name += "Scaled"
        return op_name

    @property
    def input(self):
        assert self._input is not None, f"{self.name} has not being built yet."
        return self._input[0] if len(self._input) == 1 else self._input

    @property
    def output(self):
        assert self._output is not None, f"{self.name} has not being built yet."
        return self._output

    @property
    def weights(self):
        return self._weights

    def _load_attributes(self, **kwargs):
        attrs = []
        for key, value in kwargs.items():
            # Convert each non None value in an AttributeProto
            if value is not None:
                value = to_field(key, value)
                attrs.append(value)
        return attrs

    @staticmethod
    def build_subgraph(op_type):
        """Define the subgraph

        Args:
            op_type (str): operation type to build

        Returns:
            list of NodeProto: the operation sequence.
        """
        raise NotImplementedError("Child must implement this function")

    def _add_weight(self, name, value=[], dtype="float32"):
        """Add a new weight into the object.

        Note:
            Weights have to be created on child in __init__.
        """
        self._weights[name] = np.array(value, dtype)

    def set_weight(self, name, value):
        """Set a weights that can be extracted from the float model

        Args:
            name (str): the weight to modify
            value (np.ndarray): the new value
        """
        assert isinstance(value, np.ndarray), f"Expected {value} is a numpy array."
        if name not in self.weights:
            raise ValueError(f"{self.name} ({self.base_name}) does not recognize '{name}'. "
                             f"Availables: {list(self.weights)}")
        if value.dtype != self.weights[name].dtype:
            raise ValueError(f"{self.base_name}/{name} does not match with expected type "
                             f"({self._weights[name].dtype}). Receives {value.dtype}")
        self._weights[name] = value

    def __build__(self, *input_tensor_shapes, downscale=True):
        """Build weights and compute the output shape

        Args:
            *input_tensor_shapes (tuple): the input shapes and types
            downscale (bool, optional): whether to apply downscale operation. Defaults to True.

        Returns:
            tuple: the output shape and type
        """
        raise NotImplementedError("Child must implement this function")

    def build(self, *inputs_vi, out_name=None, downscale=True):
        """Build the layer in several steps:

        1. Build extra weights, needed at quantization time.
        2. Check weights integrity given the input shape.
        3. Compute output shape.

        Args:
            inputs_vi (list of ValueInfoProto): list of inputs value info.
            out_name (str, optional): the output tensor name. Defaults to None.
            downscale (bool, optional): whether to apply downscale operation,
                which will change the output type. Defaults to True.
        """
        assert all(isinstance(x, ValueInfoProto) for x in inputs_vi)
        # Replace empty name
        if not self.name:
            self.name = str(uuid.uuid4())
        # Convert ValueInfoProto into TensorShape
        input_ts = [value_info_to_tensor_shape(x) for x in inputs_vi]
        if len(inputs_vi) > 0:
            self._input = inputs_vi
        output_ts = self.__build__(*input_ts, downscale=downscale)
        self._output = make_tensor_value_info(out_name or self.name,
                                              elem_type=np_dtype_to_tensor_dtype(output_ts.dtype),
                                              shape=output_ts.shape)
        # Special weights: each qlayer must have an output scale and (potentially) a zero point.
        # But the zero point type may change depending on the layer type.
        # That is why we add it only if child did not do it
        scale_zp_shape = output_ts.shape[1]
        self._add_weight("scale", value=np.ones(scale_zp_shape), dtype="float64")
        if "zero_point" not in self.weights:
            self._add_weight("zero_point", value=np.zeros(scale_zp_shape), dtype="int8")

    def __quantize__(self, *qlayers, out_tensor_range, force_fp=False):
        """Build weights and compute the output scale

        Args:
            qlayers (list of OnnxLayer): the input layers. Input scales and zero points
                will be deduced from these.
            out_tensor_range (TensorData): the ranges computed by calibration.
            force_fp (bool, optional): whether to force output scale as a power-of-two.
                Defaults to False.

        Returns:
            tuple: quantized weights and output scale
        """
        raise NotImplementedError("Child must implement this function")

    def quantize(self, *qlayers, out_tensor_range, force_fp=False, downscale=True):
        """Quantize the float weights given a set of input scales and zero points.

        Args:
            qlayers (list of OnnxLayer): the input layers. Input scales and zero points
                will be deduced from these.
            out_tensor_range (TensorData): the ranges computed by calibration.
            force_fp (bool, optional): whether to force output scale as a power-of-two.
                Defaults to False.
            downscale (bool, optional): whether to apply downscale operation,
                which will change the output type. Defaults to True.

        Returns:
            NodeProto, list of TensorProto: serialized objects to build the ONNX graph.
        """
        if self._output is None or self._input is None:
            # Build the layer if required
            input_ts = [qly.output for qly in qlayers]
            self.build(*input_ts, downscale=downscale)
        # Quantize weights
        qweights, output_scale = self.__quantize__(*qlayers,
                                                   out_tensor_range=out_tensor_range,
                                                   force_fp=force_fp)
        # Save output scale to be recovered for next qlayer
        self.set_weight("scale", output_scale)
        # Return ONNX node and weights
        inputs = [ts.name for ts in self._input] + list(qweights)
        onnx_node = self.make_node(inputs, [self.output.name])
        onnx_weights = array_to_tp(**qweights)
        # Although output scale is not used in the operation chain, we store it as
        # an attribute to allow us to dequantize the output at any time.
        onnx_node.attribute.append(to_field("scale", self.weights["scale"]))
        return onnx_node, onnx_weights

    def make_node(self, inputs, outputs):
        """Return the NodeProto, setting the attributes.

        Args:
            inputs (list of str): list of input names.
            outputs (list of str): list of output names.

        Returns:
            NodeProto: the corresponding node.
        """
        # Build the subgraph (implemented in derived classes) and register subgraph
        # to make it available, unless previously registered already
        nodes = self.build_subgraph(self.op_type)
        inputs_fn, outputs_fn, attributes_fn = infer_function_parameters(nodes)
        func = make_function(domain=DOMAIN,
                             fname=self.op_type,
                             inputs=inputs_fn,
                             outputs=outputs_fn,
                             nodes=nodes,
                             opset_imports=[make_opsetid("", onnx_opset_version())],
                             attributes=attributes_fn)
        register_new_subgraph(func)

        # Return the node with corresponding attributes
        node = make_node(self.op_type, inputs, outputs, self.name, domain=DOMAIN)
        consume_attrs = [attr for attr in self.attribute if attr.name in func.attribute]
        node.attribute.extend(consume_attrs)
        return node



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["fuse_relu_clip_nodes"]

import numpy as np
import onnx.helper
import onnx.numpy_helper

from ..model import ONNXModel


def _find_clip_relu_sequence(model):
    for node in model.nodes():
        node_outbounds = model.get_children(node)

        # Find candidate node and get the type of the next node if there is one
        if node.op_type in ("Clip", "Relu") and len(node_outbounds) == 1:
            outbound = node_outbounds[0]
            if outbound.op_type in ("Clip", "Relu"):
                return [node, outbound]
    return None


def _get_min_max_values(model, node):
    if node.op_type == "Relu":
        node_min = np.array(0.0, dtype=np.float32)
        node_max = np.array(np.inf, dtype=np.float32)
    else:
        # We add empty min/max if they are not present
        if len(node.input) == 1:
            node.input.extend(["", ""])
        elif len(node.input) == 2:
            node.input.append("")

        node_min = np.array(-np.inf, dtype=np.float32) if node.input[1] == "" \
            else model.get_variable(node.input[1])

        node_max = np.array(np.inf, dtype=np.float32) if node.input[2] == "" \
            else model.get_variable(node.input[2])

    return node_min, node_max


def _set_tensor_proto_value(model, node, value, suffix):
    tp_name = f"{node.input[0]}/{suffix}"
    new_node_tp = onnx.numpy_helper.from_array(value, tp_name)
    node_input_index = 1 if suffix == "min" else 2

    if (node_tp := model.get_initializer(tp_name)) is not None:
        # update tensor proto
        node_tp.CopyFrom(new_node_tp)
    else:
        # Create new tensor proto
        model.initializer().append(new_node_tp)

    node.input[node_input_index] = tp_name


def fuse_relu_clip_nodes(model):
    """Fuses successives Clip/ReLU nodes

    Args:
        model (ONNXModel): The ONNX model to be processed.
    """
    assert isinstance(model, ONNXModel)

    # We consider successively the following sequences of nodes:
    # Clip > Clip or Clip > ReLU or ReLU > Clip or ReLU > ReLU
    while True:
        # Find Clip/ReLU sequences
        sequence = _find_clip_relu_sequence(model)

        if sequence is None:
            break  # No more sequence

        # find min max of each node of the sequence
        min1, max1 = _get_min_max_values(model, sequence[0])
        min2, max2 = _get_min_max_values(model, sequence[1])

        # compute new min max values
        new_min = np.maximum(min1, min2)
        new_max = np.minimum(max1, max2)

        # create new node
        # In case both nodes are ReLUs, we create a ReLU node
        if sequence[0].op_type == "Relu" and sequence[1].op_type == "Relu":
            new_node = onnx.helper.make_node("Relu", inputs=sequence[0].input,
                                             outputs=sequence[1].output)
        else:
            new_node = onnx.helper.make_node("Clip", inputs=[sequence[0].input[0], "", ""],
                                             outputs=sequence[1].output)

            if not np.isinf(new_min):
                _set_tensor_proto_value(model, new_node, new_min, "min")

            if not np.isinf(new_max):
                _set_tensor_proto_value(model, new_node, new_max, "max")

        # Remove Unused value_info which is the output of the first node of the sequence
        value_info_to_remove = model.find_value_info_by_name(sequence[0].output[0])
        model.graph().value_info.remove(value_info_to_remove)

        # Remove the sequence and add the new node
        model.remove_nodes(sequence)
        model.add_node(new_node)

    # As we add new nodes, we need to topologically sort the model graph
    model.topological_sort()
    model.clean_initializers()
    model.check_model()



from .keras_quantization_error import (
    quantization_error as keras_layer_quantization_error,
    cumulative_quantization_error as keras_cumulative_quantization_error)
from .onnx_quantization_error import (
    quantization_error as onnx_node_quantization_error,
    cumulative_quantization_error as onnx_cumulative_quantization_error)
